"""
    domonic.terminal
    ====================================
    - call command line functions in python 3
"""
import subprocess
import os

from domonic.javascript import window


class TerminalException(Exception):
    """ raised if the terminal throws an exception """

    def __init__(self, error, message="An error message was recieved from terminal"):
        print(error, message)
        self.message = message
        super().__init__(self.message)


class command():
    """ wrapper class for all terminal commands """

    @staticmethod
    def run(cmd: str):
        """run

        runs any command on the terminal

        Args:
            cmd (str): The command you want to run on the terminal

        Returns:
            str: the response as a string
        """
        returned_output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
        return returned_output.decode("utf-8")

    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs
        self.params = ''.join([each.__str__() for each in args])
        self.result = ""

        if not hasattr(self, 'first_run'):
            self.first_run = True
        else:
            self.first_run = False

        self.run_command()

    def __call__(self, *args, **kwargs):
        self.run_command()
        return self.result

    def run_command(self):
        if hasattr(self, 'wait'):
            self.has_wait = True
            if self.first_run is True:
                self.first_run = False
                return
        else:
            self.has_wait = False

        if hasattr(self, 'iterable'):
            if self.first_run is True:
                self.first_run = False
                return

        try:
            cmd = f"{self.name} {self.params}"
            cmd = cmd.strip()

            '''
            # TODO - after doing all this. think i may have decided to go for 2 base commands instead
            # one iterable and one normal. its still hidden from user as just a different inhereted command
            # but would keep both simpler and can have more types. so will have to refactor all this again
            # for now this behaves how i want despite the _new_ hack and double call on this command
            '''
            if self.has_wait is not True:
                returned_output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
                self.result = returned_output.decode("utf-8")
            else:
                self.result = 'PING FAIL'

                def kill_switch(proc):
                    proc.kill()
                    return

                # try:
                proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
                intId = window.setInterval(3000, kill_switch, proc)
                self.result = proc.stdout.readlines()
                window.clearInterval(intId)
                # except subprocess.TimeoutExpired as e:
                # self.result = returned_output.decode("utf-8")
                # print('process ran too long')

        except subprocess.CalledProcessError as e:
            print(e.output)
            self.result = e.output
            raise TerminalException(e, self.result)

        return

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance.__init__(*args, **kwargs)
        if hasattr(instance, 'iterable'):
            is_iterable = instance.iterable
            if is_iterable:
                return instance

        return str(instance.result)

    def __str__(self) -> str:
        return str(self.result)

    def __getitem__(self, index):
        return self.result.splitlines()[index]


class cd(command):
    """
    NOTE - 'cd' does not run on terminal
    - cd is pointless as session opens and closes
    - so is overridden to change dirs via pure python
    """
    def run_command(self):
        os.chdir(self.params)


# tested --
ls = type('ls', (command,), {'name': 'ls', 'iterable': True})
rm = type('rm', (command,), {'name': 'rm'})
ln = type('ln', (command,), {'name': 'ln'})
mkdir = type('mkdir', (command,), {'name': 'mkdir'})
rmdir = type('rmdir', (command,), {'name': 'rmdir'})
touch = type('touch', (command,), {'name': 'touch'})
df = type('df', (command,), {'name': 'df', 'iterable': True})
du = type('du', (command,), {'name': 'du', 'iterable': True})
ps = type('ps', (command,), {'name': 'ps', 'iterable': True})
cat = type('cat', (command,), {'name': 'cat', 'iterable': True})
# --

mv = type('mv', (command,), {'name': 'mv'})
cp = type('cp', (command,), {'name': 'cp'})
rm = type('rm', (command,), {'name': 'rm'})
chmod = type('chmod', (command,), {'name': 'chmod'})
chown = type('chown', (command,), {'name': 'chown'})

users = type('users', (command,), {'name': 'users', 'iterable': True})
useradd = type('useradd', (command,), {'name': 'useradd'})
userdel = type('userdel', (command,), {'name': 'userdel'})
groups = type('groups', (command,), {'name': 'groups'})
groupadd = type('groupadd', (command,), {'name': 'groupadd'})
groupdel = type('groupdel', (command,), {'name': 'groupdel'})

uniq = type('uniq', (command,), {'name': 'uniq'})
sort = type('sort', (command,), {'name': 'sort'})
diff = type('diff', (command,), {'name': 'diff'})
# cmp = type('cmp', (command,), {'name': 'cmp'})

ssh = type('ssh', (command,), {'name': 'ssh'})
scp = type('scp', (command,), {'name': 'scp'})

more = type('more', (command,), {'name': 'more'})
less = type('less', (command,), {'name': 'less'})
# vi = type('vi', (command,), {'name': 'vi'})
# emacs = type('emacs', (command,), {'name': 'emacs'})
head = type('head', (command,), {'name': 'head', 'iterable': True})
tail = type('tail', (command,), {'name': 'tail', 'iterable': True})

uptime = type('uptime', (command,), {'name': 'uptime'})
bash = type('bash', (command,), {'name': 'bash'})
shutdown = type('shutdown', (command,), {'name': 'shutdown'})
reboot = type('reboot', (command,), {'name': 'reboot'})
# open = type('open', (command,), {'name': 'open'})
killall = type('killall', (command,), {'name': 'killall'})
mkfile = type('mkfile', (command,), {'name': 'mkfile'})
ifconfig = type('ifconfig', (command,), {'name': 'ifconfig', 'iterable': True})
ipconfig = type('ipconfig', (command,), {'name': 'ipconfig', 'iterable': True})
finger = type('finger', (command,), {'name': 'finger'})
passwd = type('passwd', (command,), {'name': 'passwd'})
whoami = type('whoami', (command,), {'name': 'whoami'})
# help = type('help', (command,), {'name': 'help'})


class history(command):
    def run_command(self):
        self.iterable = True
        try:
            # history is empty for non interactive sessions so read file
            from os.path import join, expanduser
            with open(join(expanduser('~'), '.bash_history'), 'r') as f:
                self.result = ''.join(f.readlines())
        except Exception as e:
            print('failed to get history:', e)
            self.result = ''


ping = type('ping', (command,), {'name': 'ping', 'wait': True, 'iterable': True})  # < TODO - need to stream feedback
man = type('man', (command,), {'name': 'man'})
find = type('find', (command,), {'name': 'find'})
awk = type('awk', (command,), {'name': 'awk'})
grep = type('grep', (command,), {'name': 'grep'})
cut = type('cut', (command,), {'name': 'cut'})
sed = type('sed', (command,), {'name': 'sed'})
apt = type('apt', (command,), {'name': 'apt'})
gunzip = type('gunzip', (command,), {'name': 'gunzip'})
tar = type('tar', (command,), {'name': 'tar'})
gzip = type('gzip', (command,), {'name': 'gzip'})
cron = type('cron', (command,), {'name': 'cron'})
crontab = type('crontab', (command,), {'name': 'crontab'})
xargs = type('xargs', (command,), {'name': 'xargs'})
nautilus = type('nautilus', (command,), {'name': 'nautilus'})  # Linux only?
date = type('date', (command,), {'name': 'date'})
cal = type('cal', (command,), {'name': 'cal'})
bc = type('bc', (command,), {'name': 'bc'})

rsync = type('rsync', (command,), {'name': 'rsync'})

# https://github.com/agarrharr/awesome-cli-apps
# - NOTE - these ones may need installing if you haven't already
git = type('git', (command,), {'name': 'git'})
wget = type('wget', (command,), {'name': 'wget'})
nmap = type('nmap', (command,), {'name': 'nmap'})
nohup = type('nohup', (command,), {'name': 'nohup'})
python = type('python', (command,), {'name': 'python'})
npm = type('npm', (command,), {'name': 'npm'})
cowsay = type('cowsay', (command,), {'name': 'cowsay'})
pip = type('pip', (command,), {'name': 'pip'})  # TODO - change to 'python3 -m pip'? # TODO - both iterable and not
say = type('say', (command,), {'name': 'say'})  # Mac only?
gcc = type('gcc', (command,), {'name': 'gcc'})
jq = type('jq', (command,), {'name': 'jq'})
curl = type('curl', (command,), {'name': 'curl'})
ffmpeg = type('ffmpeg', (command,), {'name': 'ffmpeg'})
convert = type('convert', (command,), {'name': 'convert'})  # imagemagik
figlet = type('figlet', (command,), {'name': 'figlet'})
banner = type('banner', (command,), {'name': 'banner'})
# imagemagik
# asciiview
# toilet
# figlet
# youtube-dl = type('youtube-dl', (command,), {'name': 'youtube-dl'}) # TODO - illegal chars
# tree = type('tree', (command,), {'name': 'tree'})
# vim = type('vim', (command,), {'name': 'vim'})
# mail = type('mail', (command,), {'name': 'mail'})
# vncserver = type('vncserver', (command,), {'name': 'vncserver'})
# tmux

# https://en.wikipedia.org/wiki/List_of_Unix_commands
# admin = type('admin', (command,), {'name': 'admin'})
ar = type('ar', (command,), {'name': 'ar'})
asa = type('asa', (command,), {'name': 'asa'})
at = type('at', (command,), {'name': 'at'})
basename = type('basename', (command,), {'name': 'basename'})
batch = type('batch', (command,), {'name': 'batch'})
bg = type('bg', (command,), {'name': 'bg'})
cc = type('cc', (command,), {'name': 'cc'})
cflow = type('cflow', (command,), {'name': 'cflow'})
chgrp = type('chgrp', (command,), {'name': 'chgrp'})
cksum = type('cksum', (command,), {'name': 'cksum'})
comm = type('comm', (command,), {'name': 'comm'})
compress = type('compress', (command,), {'name': 'compress'})
csplit = type('csplit', (command,), {'name': 'csplit'})
ctags = type('ctags', (command,), {'name': 'ctags'})
cxref = type('cxref', (command,), {'name': 'cxref'})
dd = type('dd', (command,), {'name': 'dd'})
delta = type('delta', (command,), {'name': 'delta'})
dirname = type('dirname', (command,), {'name': 'dirname'})
ed = type('ed', (command,), {'name': 'ed'})
env = type('env', (command,), {'name': 'env'})
ex = type('ex', (command,), {'name': 'ex'})
expand = type('expand', (command,), {'name': 'expand'})
expr = type('expr', (command,), {'name': 'expr'})
fc = type('fc', (command,), {'name': 'fc'})
fg = type('fg', (command,), {'name': 'fg'})
file = type('file', (command,), {'name': 'file'})
fold = type('fold', (command,), {'name': 'fold'})
fort77 = type('fort77', (command,), {'name': 'fort77'})
fuser = type('fuser', (command,), {'name': 'fuser'})
gencat = type('gencat', (command,), {'name': 'gencat'})
get = type('get', (command,), {'name': 'get'})
getconf = type('getconf', (command,), {'name': 'getconf'})
getopts = type('getopts', (command,), {'name': 'getopts'})
iconv = type('iconv', (command,), {'name': 'iconv'})
# id = type('id', (command,), {'name': 'id'})
ipcrm = type('ipcrm', (command,), {'name': 'ipcrm'})
ipcs = type('ipcs', (command,), {'name': 'ipcs'})
join = type('join', (command,), {'name': 'join'})
lex = type('lex', (command,), {'name': 'lex'})
link = type('link', (command,), {'name': 'link'})
locale = type('locale', (command,), {'name': 'locale'})
localedef = type('localedef', (command,), {'name': 'localedef'})
logger = type('logger', (command,), {'name': 'logger'})
logname = type('logname', (command,), {'name': 'logname'})
lp = type('lp', (command,), {'name': 'lp'})
m4 = type('m4', (command,), {'name': 'm4'})
mailx = type('mailx', (command,), {'name': 'mailx'})
make = type('make', (command,), {'name': 'make'})
mesg = type('mesg', (command,), {'name': 'mesg'})
mkdir = type('mkdir', (command,), {'name': 'mkdir'})
mkfifo = type('mkfifo', (command,), {'name': 'mkfifo'})
more = type('more', (command,), {'name': 'more'})
newgrp = type('newgrp', (command,), {'name': 'newgrp'})
nice = type('nice', (command,), {'name': 'nice'})
nl = type('nl', (command,), {'name': 'nl'})
nm = type('nm', (command,), {'name': 'nm'})
od = type('od', (command,), {'name': 'od'})
paste = type('paste', (command,), {'name': 'paste'})
patch = type('patch', (command,), {'name': 'patch'})
pathchk = type('pathchk', (command,), {'name': 'pathchk'})
pax = type('pax', (command,), {'name': 'pax'})
pr = type('pr', (command,), {'name': 'pr'})
prs = type('prs', (command,), {'name': 'prs'})
qalter = type('qalter', (command,), {'name': 'qalter'})
qdel = type('qdel', (command,), {'name': 'qdel'})
qhold = type('qhold', (command,), {'name': 'qhold'})
qmove = type('qmove', (command,), {'name': 'qmove'})
qmsg = type('qmsg', (command,), {'name': 'qmsg'})
qrerun = type('qrerun', (command,), {'name': 'qrerun'})
qrls = type('qrls', (command,), {'name': 'qrls'})
qselect = type('qselect', (command,), {'name': 'qselect'})
qsig = type('qsig', (command,), {'name': 'qsig'})
qstat = type('qstat', (command,), {'name': 'qstat'})
qsub = type('qsub', (command,), {'name': 'qsub'})
read = type('read', (command,), {'name': 'read'})
renice = type('renice', (command,), {'name': 'renice'})
rmdel = type('rmdel', (command,), {'name': 'rmdel'})
sact = type('sact', (command,), {'name': 'sact'})
sccs = type('sccs', (command,), {'name': 'sccs'})
sed = type('sed', (command,), {'name': 'sed'})
sh = type('sh', (command,), {'name': 'sh'})
sleep = type('sleep', (command,), {'name': 'sleep'})
split = type('split', (command,), {'name': 'split'})
strings = type('strings', (command,), {'name': 'strings'})
strip = type('strip', (command,), {'name': 'strip'})
stty = type('stty', (command,), {'name': 'stty'})
tabs = type('tabs', (command,), {'name': 'tabs'})
talk = type('talk', (command,), {'name': 'talk'})
tee = type('tee', (command,), {'name': 'tee'})
test = type('test', (command,), {'name': 'test'})
time = type('time', (command,), {'name': 'time'})
tput = type('tput', (command,), {'name': 'tput'})
tr = type('tr', (command,), {'name': 'tr'})
true = type('true', (command,), {'name': 'true'})
tsort = type('tsort', (command,), {'name': 'tsort'})
tty = type('tty', (command,), {'name': 'tty'})
uname = type('uname', (command,), {'name': 'uname'})
uncompress = type('uncompress', (command,), {'name': 'uncompress'})
unexpand = type('unexpand', (command,), {'name': 'unexpand'})
unget = type('unget', (command,), {'name': 'unget'})
unlink = type('unlink', (command,), {'name': 'unlink'})
uucp = type('uucp', (command,), {'name': 'uucp'})
uudecode = type('uudecode', (command,), {'name': 'uudecode'})
uuencode = type('uuencode', (command,), {'name': 'uuencode'})
uustat = type('uustat', (command,), {'name': 'uustat'})
uux = type('uux', (command,), {'name': 'uux'})
val = type('val', (command,), {'name': 'val'})
wc = type('wc', (command,), {'name': 'wc'})  # < TODO - stream. plus figure out piping
what = type('what', (command,), {'name': 'what'})
who = type('who', (command,), {'name': 'who'})
write = type('write', (command,), {'name': 'write'})
yacc = type('yacc', (command,), {'name': 'yacc'})
zcat = type('zcat', (command,), {'name': 'zcat'})

# bash builtins
alias = type('alias', (command,), {'name': 'alias'})
bg = type('bg', (command,), {'name': 'bg'})
bind = type('bind', (command,), {'name': 'bind'})
# break = type('break', (command,), {'name': 'break'})
builtin = type('builtin', (command,), {'name': 'builtin'})
caller = type('caller', (command,), {'name': 'caller'})
# command = type('command', (command,), {'name': 'command'})
compgen = type('compgen', (command,), {'name': 'compgen'})
complete = type('complete', (command,), {'name': 'complete'})
compopt = type('compopt', (command,), {'name': 'compopt'})
# continue = type('continue', (command,), {'name': 'continue'})
declare = type('declare', (command,), {'name': 'declare'})
dirs = type('dirs', (command,), {'name': 'dirs'})
disown = type('disown', (command,), {'name': 'disown'})
echo = type('echo', (command,), {'name': 'echo'})
enable = type('enable', (command,), {'name': 'enable'})
# eval = type('eval', (command,), {'name': 'eval'})
# exec = type('exec', (command,), {'name': 'exec'})
exit = type('exit', (command,), {'name': 'exit'})
export = type('export', (command,), {'name': 'export'})
# false = type('false', (command,), {'name': 'false'})
fc = type('fc', (command,), {'name': 'fc'})
fg = type('fg', (command,), {'name': 'fg'})
getopts = type('getopts', (command,), {'name': 'getopts'})
# hash = type('hash', (command,), {'name': 'hash'})
# help = type('help', (command,), {'name': 'help'})

jobs = type('jobs', (command,), {'name': 'jobs'})
kill = type('kill', (command,), {'name': 'kill'})
let = type('let', (command,), {'name': 'let'})
local = type('local', (command,), {'name': 'local'})
logout = type('logout', (command,), {'name': 'logout'})
mapfile = type('mapfile', (command,), {'name': 'mapfile'})
popd = type('popd', (command,), {'name': 'popd'})
printf = type('printf', (command,), {'name': 'printf'})
pushd = type('pushd', (command,), {'name': 'pushd'})
pwd = type('pwd', (command,), {'name': 'pwd'})
read = type('read', (command,), {'name': 'read'})
readarray = type('readarray', (command,), {'name': 'readarray'})
readonly = type('readonly', (command,), {'name': 'readonly'})
# return = type('return', (command,), {'name': 'return'})
# set = type('set', (command,), {'name': 'set'})
shift = type('shift', (command,), {'name': 'shift'})
shopt = type('shopt', (command,), {'name': 'shopt'})
source = type('source', (command,), {'name': 'source'})
suspend = type('suspend', (command,), {'name': 'suspend'})
# test = type('test', (command,), {'name': 'test'})
times = type('times', (command,), {'name': 'times'})
trap = type('trap', (command,), {'name': 'trap'})
# true = type('true', (command,), {'name': 'true'})
# type = type('type', (command,), {'name': 'type'})
typeset = type('typeset', (command,), {'name': 'typeset'})
ulimit = type('ulimit', (command,), {'name': 'ulimit'})
umask = type('umask', (command,), {'name': 'umask'})
unalias = type('unalias', (command,), {'name': 'unalias'})
unset = type('unset', (command,), {'name': 'unset'})
wait = type('wait', (command,), {'name': 'wait'})


# https://ss64.com/bash/
# PIPE etc?



"""
    domonic.cmd
    ====================================
    - call cmd commands from python 3
"""
import subprocess
import os

from domonic.javascript import window

# create a mapping of equivelent commands to call on terminal if user OS is not windows
equivelents_map = {
    "dir": "ls",
    "move": "mv",
    "copy": "cp",
    "erase": "rm",
    "comp": "diff",
    # "type_":"cat",
    # "open":"open",
    # "edit":"vim",
    # "view":"less",
    # "search":"grep",
    # "replace":"sed"
}


class CmdException(Exception):
    """ raised if cmd throws an exception """

    def __init__(self, error, message: str="An error message was recieved from cmd"):
        print(error, message)
        self.message = message
        super().__init__(self.message)


class Cmdcommand():
    """ wrapper class for all cmd commands """

    @staticmethod
    def run(cmd: str) -> str:
        """[runs any command on the cmd]

        Args:
            cmd (str): The command you want to run on cmd

        Returns:
            str: the response as a string
        """
        returned_output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
        return returned_output.decode("utf-8")

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.params = ''.join([each.__str__() for each in args])
        self.result = ""

        if not hasattr(self, 'first_run'):
            self.first_run = True
        else:
            self.first_run = False

        # TODO - call terminal equiv command if not on windows
        # from sys import platform
        # if platform == "linux" or platform == "linux2":
        #     print("This is linux")
        # elif platform == "darwin":
        #    print("This is mac")
        # elif platform == "win32":
        #    print("This is windows")

        self.run_command()

    def __call__(self, *args, **kwargs):
        self.run_command()
        return self.result

    def run_command(self):
        if hasattr(self, 'wait'):
            self.has_wait = True
            if self.first_run is True:
                self.first_run = False
                return
        else:
            self.has_wait = False

        if hasattr(self, 'iterable'):
            if self.first_run is True:
                self.first_run = False
                return

        try:
            cmd = f"{self.name} {self.params}"
            cmd = cmd.strip()

            '''
            # TODO - after doing all this. think i may have decided to go for 2 base commands instead
            # one iterable and one normal. its still hidden from user as just a different inhereted command
            # but would keep both simpler and can have more types. so will have to refactor all this again
            # for now this behaves how i want despite the _new_ hack and double call on this command
            '''
            if self.has_wait is not True:
                returned_output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
                self.result = returned_output.decode("utf-8")
            else:
                self.result = 'PING FAIL'

                def kill_switch(proc):
                    proc.kill()
                    return

                # try:
                proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
                intId = window.setInterval(3000, kill_switch, proc)
                self.result = proc.stdout.readlines()
                window.clearInterval(intId)
                # except subprocess.TimeoutExpired as e:
                # self.result = returned_output.decode("utf-8")
                # print('process ran too long')

        except subprocess.CalledProcessError as e:
            print(e.output)
            self.result = e.output
            raise CmdException(e, self.result)

        return

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance.__init__(*args, **kwargs)
        if hasattr(instance, 'iterable'):
            is_iterable = instance.iterable
            if is_iterable:
                return instance

        return str(instance.result)

    def __str__(self):
        return str(self.result)

    def __getitem__(self, index):
        return self.result.splitlines()[index]

    # def __repr__(self):
    #     return f"{self.__class__.__name__}({self.params})"

    def __iter__(self):
        return iter(self.result.splitlines())

    # def __add__(self, other):
    #     return str(self.result) + str(other)

    # def __radd__(self, other):
    #     return str(other) + str(self.result)

    # def __call__(self, *args, **kwargs):
    #     self.run_command()
    #     return self.result

    # def __getattr__(self, attr):
    #     if attr in equivelents_map:
    #         return Cmdcommand(equivelents_map[attr], *self.args, **self.kwargs)

    # def __del__(self):
    #     os.chdir("..")
    #     return


class cd(Cmdcommand):
    """
    NOTE - 'cd' does not run on cmd
    - cd is pointless as session opens and closes
    - so is overridden to change dirs via pure python
    """
    def run_command(self):
        os.chdir(self.params)


# tested --
dir = type('dir', (Cmdcommand,), {'name': 'dir', 'iterable': True})  #: list directory content
erase = type('erase', (Cmdcommand,), {'name': 'erase'})
# del = type('del', (command,), {'name': 'del'})
mkdir = type('mkdir', (Cmdcommand,), {'name': 'mkdir'})  #: create a new directory
rmdir = type('rmdir', (Cmdcommand,), {'name': 'rmdir'})  #: delete directory
copy = type('copy', (Cmdcommand,), {'name': 'copy'})  #: copy files

fsutil = type('fsutil', (Cmdcommand,), {'name': 'fsutil'})
fc = type('fc', (Cmdcommand,), {'name': 'fc'})  # compare files and display the differences


class touch(Cmdcommand):
    def run_command(self):
        self.iterable = True
        try:
            cmd = f"fsutil file createnew {self.params} 0"
            Cmdcommand.run(cmd)
        except Exception as e:
            # print('failed to touch:', e)
            self.result = ''


getmac = type('getmac', (Cmdcommand,), {'name': 'getmac'})  #: display MAC address
ipconfig = type('ipconfig', (Cmdcommand,), {'name': 'ipconfig'})  #: display IP network settings
shutdown = type('shutdown', (Cmdcommand,), {'name': 'shutdown'})  #: shutdown the computer. (/s), triggers a restart (/r), or logs the user out (/l).
# date = type('date', (Cmdcommand,), {'name': 'date'})  #: show/set date - TODO seemed to hang. needs a param
echo = type('echo', (Cmdcommand,), {'name': 'echo'})  #: text output
hostname = type('hostname', (Cmdcommand,), {'name': 'hostname'})  #: display host name
# time = type('time', (Cmdcommand,), {'name': 'time'})  #: display/edit the system time - TODO seemed to hang
ver = type('ver', (Cmdcommand,), {'name': 'ver'})  #: display operating system version - TODO seemed to hang
# netstat = type('netstat', (Cmdcommand,), {'name': 'netstat'})  #: display TCP/IP connections and status
ping = type('ping', (Cmdcommand,), {'name': 'ping'})  #: pings the network
# ping = type('ping', (Cmdcommand,), {'name': 'ping', 'wait': True, 'iterable': True})  # < TODO - need to stream feedback

move = type('move', (Cmdcommand,), {'name': 'move'})  #: move/rename files
rename = type('rename', (Cmdcommand,), {'name': 'rename'})  #: rename files
replace = type('replace', (Cmdcommand,), {'name': 'replace'})  #: replace files

systeminfo = type('systeminfo', (Cmdcommand,), {'name': 'systeminfo'})  #: displays computer-specific properties and configurations

attrib = type('attrib', (Cmdcommand,), {'name': 'attrib'})  #: display file attributes
# tree = type('tree', (Cmdcommand,), {'name': 'tree'})  #: display folder structure graphically - TODO - return not utf-8
type_ = type('type_', (Cmdcommand,), {'name': 'type'})  #: display content of text files
comp = type('comp', (Cmdcommand,), {'name': 'comp'})  #: compare file contents

chkdsk = type('chkdsk', (Cmdcommand,), {'name': 'chkdsk'})  #: check volumes
driverquery = type('driverquery', (Cmdcommand,), {'name': 'driverquery'})  #: display installed devices and their properties
vol = type('vol', (Cmdcommand,), {'name': 'vol'})  #: show volume description and serial numbers of the HDDs
gpresult = type('gpresult', (Cmdcommand,), {'name': 'gpresult'})  #: display group policies

# ssh = type('ssh', (Cmdcommand,), {'name': 'ssh'})

chdir = type('chdir', (Cmdcommand,), {'name': 'chdir'})  # : show current dir or can switch dir
# clip = type('clip', (Cmdcommand,), {'name': 'clip'})  # : Forwards the result of a command to the clipboard

# find = type('find', (Cmdcommand,), {'name': 'find'})
whoami = type('whoami', (Cmdcommand,), {'name': 'whoami'})  #: information about the current user. /GROUP parameter

logoff = type('logoff', (Cmdcommand,), {'name': 'logoff'})  #: Logs the user out of Windows.
mrinfo = type('mrinfo', (Cmdcommand,), {'name': 'mrinfo'})  #: Provides information on the router
tasklist = type('tasklist', (Cmdcommand,), {'name': 'tasklist'})  #: Lists all running processes

# cmd = type('cmd', (Cmdcommand,), {'name': 'cmd'})  #: start command prompt - NOTE hangs
title = type('title', (Cmdcommand,), {'name': 'title'})  #: Changes the title of the command prompt
tzutil = type('tzutil', (Cmdcommand,), {'name': 'tzutil'})  #: Displays the currently set time zone (/g) or changes it (/s)

# bitsadmin = type('bitsadmin', (Cmdcommand,), {'name': 'bitsadmin'})  #: Allows you to run commands on a remote computer
# chcp = type('chcp', (Cmdcommand,), {'name': 'chcp'})  #: Change the locale of the command prompt
# cls = type('cls', (Cmdcommand,), {'name': 'cls'})  #: Clears the screen
# deluser = type('deluser', (Cmdcommand,), {'name': 'deluser'})  #: Deletes a user
# dir = type('dir', (Cmdcommand,), {'name': 'dir'})  #: Displays the contents of a directory
# dlls = type('dlls', (Cmdcommand,), {'name': 'dlls'})  #: Lists the DLLs loaded by an executable
# echo_ = type('echo_', (Cmdcommand,), {'name': 'echo_'})  #: text output
# echo = type('echo', (Cmdcommand,), {'name': 'echo'})  #: text output
# ed = type('ed', (Cmdcommand,), {'name': 'ed'})  #: Edit a file
# edlin = type('edlin', (Cmdcommand,), {'name': 'edlin'})  #: Edit a file
# exit = type('exit', (Cmdcommand,), {'name': 'exit'})  #: Exits the command prompt
# find = type('find', (Cmdcommand,), {'name': 'find'})  #: Find a file or folder
# findstr = type('findstr', (Cmdcommand,), {'name': 'findstr'})  #: Find a string in files or folders
# format = type('format', (Cmdcommand,), {'name': 'format'})  #: Format a file
# gpedit = type('gpedit', (Cmdcommand,), {'name': 'gpedit'})  #: Edit the Group Policy
# help = type('help', (Cmdcommand,), {'name': 'help'})  #: Displays help files
# help_ = type('help_', (Cmdcommand,), {'name': 'help_'})  #: Displays help files
# help_ctx = type('help_ctx', (Cmdcommand,), {'name': 'help_ctx'})  #: Displays help files
# help_nt = type('help_nt', (Cmdcommand,), {'name': 'help_nt'})  #: Displays help files


# WINDOWS COMMANDS
'''
bitsadmin	Creates and monitors downloads and uploads
chcp	Changes the current code page
choice	Creates a selection list
clip	Forwards the result of a command to the clipboard.
color	Changes the background
command	Starts CMD.COM.	32-bit/DOS
date	Displays the current date and allows you to change it.
debug	Starts debug
doskey	Creates macros, recalls commands, and edits command input
dosshell	Opens the DOS shell
edit	Starts the MS-DOS editor
edlin	Creates and edits text files within the command prompt
fasthelp	Displays helpful information about commands
fastopen	Writes the position of a program into a specified list
find	Searches through a file or multiple files for a particular character sequence.
findstr	Finds character sequences in one or multiple files.
forcedos	Starts a program in the MS-DOS partial system
graftabl	Enables the option to use extended characters of a specific code page in graphics mode.	32-bit/DOS
graphics	Starts a program that can print graphics.	32-bit/DOS
help	Displays help text for a specific command (you can also use the /? command)
kb16	Changes the country settings of the keyboard for DOS programs
keyb	Changes the country settings of the keyboard for DOS programs
lpq	Displays the status of a printer queue for computers that use a “line Printer Daemon” (LPD)
lpr	Sends a file to a computer that uses a line printer daemon (LPD).
md	Creates a new directory on the specified path.
more	Outputs the content of a file (for example, a text file) by the page.
msg	Sends a message to another user.
nlsfunc	Provides country-specific information for language support.	32-bit/DOS
ntbackup	Runs backup services directly from the command line or as part of batch or script files.	XP
path	Creates and displays the path for searching executable files
pause	Pauses execution in batch files and scripts.
popd	Changes to the folder saved by the pushd command.
print	Prints a text file.
prompt	Changes the display of the command prompt
pushd	Saves a specific path into a script or batch file.
qbasic	Starts qbasic, a program environment based on the BASIC programming language
rd	Deletes a directory.
rem	Writes comments in batch and script files that aren’t taken into account when executing
runas	Allows a user to run commands with the rights of another user.
scandisk	Starts Microsoft ScanDisk. The program searches data carriers for errors
schtasks	Sets the execution of specified programs and commands for a specified point in time.
set	Displays environmental variables of CMD.EXE and lets you configure them
shift	Moves variables within batch files and scripts
sort	Lists out data (from a file or command) and outputs it again sorted
start	Opens a new command prompt window in which you can run a specific program or command
subst	Assigns a drive letter to a path to create a virtual drive
taskkill	Ends one or more running tasks.
time	Displays the current time and allows it to be changed.
timeout	Stops a process for a specified time.
Files
CMD command	Description	Windows version
append	Sets the path in which files will be searched for.	32-bit/DOS
assoc	Changes the program that’s linked with a particular file ending.
attrib	Changes attributes of specified files. With the parameter +R you can protect a file from changes
cipher	Displays and changes the encryption status of files and directories on NTFS partitions.
comp	Compares the content of two files or two file sets.
compact	Displays and changes the compression status of files and directories on NTFS partitions.
copy	Copies a file or multiple files to another location.
cscript	Runs scripts over the Microsoft Script Host.
del	Deletes a file or multiple files.
deltree	Deletes a directory as well as all subdirectories and files within
diantz	Compresses files without any loss (command has the same function as makecab)
diskcomp	Compares the content of two disks (not 10)/DOS
diskcopy	Copies the content of a disk to another (not 10)/DOS
endlocal	Ends the valid range of changes to batch files or scripts.
erase	Function is the same as del
exe2bin	Converts an EXE file to a BIN file.	32-bit
expand	Extracts files and folders stored in CAB files (not 64-bit XP)/DOS
extrac32	Extracts files and folders stored in CAB files.
fc	Compares two individual files or two sets of files with one another and displays the differences
forfiles	Selects one or more files and runs a command that refers to these files.
ftype	Specifies a program for opening a specific file type.
goto	Skips the execution within a batch program to a specific line (marker)
if	Represents a conditional statement and executes expressions within batch files only under certain conditions.
makecab	Compresses files without loss in CAB format (you can also use the diantz command).
mklink	Creates a symbolic link to a file. With /D you can also create connections to directories.
move	Moves a file or multiple files from one directory to another.
openfiles	Displays and separates open system files and folders.
recover	Restores readable files that were on a defective data drive.
ren	Changes the name of a particular file.
rename	Function is the same as ren
replace	Replaces the selected file or files with one or more other files.
robocopy	Allows so-called robust file copying.
rsm	Manages media on removable storage devices.
setlocal	Limits the valid range of changes to batch files or scripts.
share	Installs file sharing and file locking.	32-bit/DOS
sxstrace	Starts the WinSxs Tracing Utility, a tool for programming diagnostics.
takeown	Restores administrator access rights to a file that have been lost when reassigning a user.
undelete	Undoes the deletion of a file
verify	When enabled, checks whether files are written correctly on a data drive.
where	Finds files that match a particular search topic.
xcopy	Copies files and entire directory structures.
System
CMD command	Description	Windows version
at	Starts commands and programs at a particular time.
auditpol	Displays current monitoring policies.
backup	Creates backups of files. These can be recovered with restore (replaced by msbackup)
bcdboot	Creates and repairs start files
bcdedit	Allows users to make changes to start configuration data storage
bdehdcfg	Prepares a hard drive for BitLocker Drive Encryption.
bootcfg	Creates, edits, or displays the content of boot.ini
bootsect	Modifies the master boot code sot that it’s compatible with the Windows Boot Manager or NT Loader
cacls	Edits and displays the access control list.
chkdsk	Checks and repairs (with the parameter /R) a data drive
chkntfs	Changes or displays the data driver check at startup.
cmdkey	Can display (/list), create (/add), or delete (/delete) login information.
convert	Converts partitions from FAT/FAT32 to NTFS.
ctty	Changes the standard input and output for the system
dblspace	Creates or configures compresses drives (a newer version of the command is called drvspace)
defrag	Defragments all or only specified drives. Use /U to observe the progress.
diskpart	Manages, creates, and deletes partitions from the hard drive.
diskperf	Allows users to remotely control the disk performance counter.
diskraid	Manages RAID systems.
dism	Manages and integrates Windows images
dispdiag	Creates a file in the current directory in which you’ll find information about your display.
dosx	Starts the DOS Protected Mode Interface
driverquery	Creates a list with all installed drivers.
drvspace	Creates or configures compressed drives.
emm386	Provides DOS with more than 640 KB of RAM
esentutl	Manages databases within the extensible storage engine.
eventcreate	Creates an entry (ID and message) in an event log.
eventtriggers	Configures and displays event trigger.	XP
fdisk	Creates, deletes, and manages partitions on the hard drive.
fltmc	Allows users to manage and display filter drivers.
fondue	Installs additional Windows features.
format	Formats a drive to the file system specified by the user
hwrcomp	Compiles self-created dictionaries for handwriting recognition
hwrreg	Installs a compiled dictionary for handwriting recognition
icacls	Edits and displays the access control list.
ktmutil	Starts the kernel transaction manager.
label	Changes or deletes a drive’s label
lh	Loads a program into the high memory area (UMB) – has the same function as loadhigh
licensingdiag	Creates an XML and a CAB file that contain information on the Windows product licence
loadfix	Ensures that a program is loaded and executed above the first 64 KB of RAM.	32-bit/DOS
loadhigh	Has the same function as lh
lock	Locks a drive so that only a user-selected program can access it directly.	98/95
lodctr	Updates all registry entries that have to do with performance indicators
logman	Creates and manages event trace sessions and performance logs.
manage-bde	Configures drive encryption with BitLocker.
mem	Displays information about the RAM and indicates which programs are currently loaded in it.	32-bit/DOS
memmaker	Optimises the RAM
mode	Configures system devices – primarily on the COM or LPT port
mofcomp	Analyses files in managed object format (MOF) and adds the classes and instances to the WMI repository
mountvol	Creates and deletes mount points for drives and displays them.
msav	Starts Microsoft Antivirus
msbackup	Starts Microsoft Backup (replaces backup and restores)
mscdex	Loads the CD-ROM support for MS-DOS
msd	Starts the program Microsoft Diagnostics, with which system information can be displayed
msiexec	Starts the Windows installer
muiunattend	Starts an automatic setup process for the multilingual user interface (MUI)
netcfg	Installs the minimal operating system Microsoft Windows PE.
ocsetup	Installs additional Windows functions.	8/7/Vista
pentnt	Recognises floating point division errors
pkgmgr	Installs, uninstalls, and configures packages and functions for Windows.
pnpunattend	Automates the installation of device drivers
pnputil	Installs plug-and-play devices from the command prompt.
power	Uses the IDLE status of a processor to reduce energy consumption
powercfg	Allows the user to change the computer’s energy options and control energy conservation plans.
pwlauncher	Configures the startup options for Windows To Go with which you can boot Windows from a USB drive
qprocess	Provides information on running processes.
query	Displays the status of a particular service.
quser	Provides information on the currently logged-in users.
reagentc	Configures the Windows recovery environment
recimg	Creates a user-defined Windows image to restore the system.	8
reg	Manages the registry of the command prompt.
regini	Changes registry authorisations.
register-cimprovider	Registers a common information model provider (CIM provider) in Windows
regsvr32	Registers a DLL file in the registry.
relog	Creates new performance indicator protocols from the data in the existing protocols.
repair-bde	Repairs and decrypts defective drives that are encrypted with BitLocker.
reset	Resets a session. You can also use the rwinsta command.
restore	Restores backups that were created with the backup command (replaced by msbackup)
rwinsta	Command has the same function as reset.
sc	Manages services by connecting to the Service Controller.
scanreg	Repairs the registry and allows a backup to be created of it.	98/95
sdbinst	Applies user-defined database files (SDB).
secedit	Analyses the security settings by comparing the current configurations with templates.
setver	Sets a version number of MS-DOS that’s forwarded to a program
setx	Creates or changes environmental variable in the user of system environment.
sfc	Checks all important and protected system files.
smartdrv	Starts and manages the hard drive cache program SMARTDrive
sys	Copies system files from MS-DOS and the command interpreter to another hard drive. This makes it bootable
systeminfo	Displays information about the Windows installation, including all installed service packages.
tpmvscmgr	Creates and deletes TPM virtual smart cards.
tracerpt	Processes logs or real-time data generated during the tracing of computer programs.
typeperf	Displays performance counter data or writes it into a file.
unformat	Undoes the drive formatting done by the format command
unlock	Unlocks a drive that was locked with the lock command.	98/95
unlodctr	Deletes names as well as descriptions for extensible performance counters in the Windows registry.
vaultcmd	Creates, deletes, and displays saved registration information
vol	Displays the label and serial number of a drive
vsafe	Starts the antivirus software VSafe
vssadmin	Manages the volume shadow copy services that can be used to store different versions (snapshots) of drives.
wbadmin	Creates backups of the operating system and delivers information to the created backup copies.
wevtutil	Manages event logs and event log files.
winmgmt	Manages WMI repositories. Backups (/backup) are possible with the command, for example
winsat	Evaluates various system factors – for example, processor performance or graphical capabilities.
wmic	Starts the Windows Management Instrumentation in the command prompt.
xwizard	Registers Windows data in the form of XML files
Network
CMD command	Description	Windows version
arp	Displays and edits entries in the Address Resolution Protocol cache
atmadm	Displays information on asynchronous transfer mode (ATM).	XP
certreq	Manages and creates certificate registration requirements for certification authorities.
certutil	Manages services related to certificate authentication.
change	Changes the settings of a terminal server and can be used together with the parameters logon, port, or user
checknetisolation	Checks the network capability of apps from the Windows Store
chglogon	Enables, disables, or adjusts logins for terminal server sessions.
chgport	Displays or changes the COM pin assignment of terminal servers for DOS compatibility.
chgusr	Changes the installation mode of a terminal server.
cmstp	Installs or uninstalls profiles for the connection manager.
djoin	Creates a new computer account in the Active Directory Domain Services (AD DS).
finger	Provides information about users on remote devices using the Finger service.
ftp	Transfers data to an FTP server or from this to a PC.
gpresult	Displays information on the Group Policy.
gpupdate	Updates information on the Group Policy.
interlnk	Connects two computers via serial or parallel connection to share files or printers
intersvr	Starts an interlnk server and transfers data from one computer to another via serial or parallel connection
ipxroute	Changes and displays information on the IPX routing tables.	XP
irftp	Transfers files via infrared connection, if one is available.
iscsicli	Manages iSCSI, which enables connections via the SCSI protocol.
klist	Displays all tickets authenticated by the Kerberos service. Also enables the command to delete tickets (purge)
ksetup	Configures a connection to a Kerberos server
mount	Enables network sharing under the Network File System. (To use the command, enable NFS services)
nbtstat	Displays statistics and information on the TCP/IP connections on remote computers
net	Configures and displays network settings
net1	Configures and displays network settings
netsh	Starts the network shell, which allows for network settings to be changed on local and remote computers.
netstat	Displays statistics and information on TCP/IP connections on the local computer
nfsadmin	Manages NFS servers and clients (to be able to use the command, you first have to enable NFS services in Windows)
nltest	Displays information related to secure channels in the Active Directory Domain Services (AD DS) and tests the connections
nslookup	Sends a DNS query to a specific IP or host name on the preconfigured DNS server. You can also specify another DNS server
ntsd	Runs debugging.	XP
pathping	Provides information on forwarding and package loss when sending over a network and also specifies the latency.
qappsrv	Displays all available remote computers in the network.
qwinsta	Displays information on the open remote desktop sessions.
rasautou	Manages autodial addresses.
rasdial	Starts and ends network connections for Microsoft clients.
rcp	Copies files from a Windows computer to a server that’s running a RSDH daemon, and vice versa
rdpsign	Signs a remote desktop protocol file (RDP file)/7
rexec	Runs commands on a remote computer that’s running a Rexec daemon.	Vista/XP
route	Displays routing tables and makes it possible to change, add, or delete entries
rpcinfo	Sends a remote procedure call (RPC) to an RPC server.
rpcping	Sends a ping via remote procedure call (RPC) and checks whether a connection is possible.
rsh	Runs commands on remote computers that are running the Unix program Remote Shell (RSH)
setspn	Creates, deletes, and changes SPNs.
shadow	Monitors a session on a remote computer.
showmount	Provides information on NFS file systems (to use the command, you first have to activate NFS services in Windows)
tcmsetup	Enables or disables a client for the Telephony Application Programming Interface (TAPI)
telnet	Enables communication with another computer that also uses the telnet protocol
tftp	Enables a file exchange between the local computer and a server that supports the Trivial File Transfer Protocol (TFTP).
tlntadmn	Manages a telnet server on a local or remote computer
tracert	Tracks a data package on the way through the network to a server.
tscon	Connects the current local user session with a session on a remote computer.
tsdiscon	Ends the connection between a local user session and a session on a remote computer.XP
tskill	Ends a process on a remote computer.
tsshutdn	Shuts down or restarts a remote terminal server.
umount	Removes mounted network file system drives.
w32tm	Manages the Windows time service that synchronises dates and times on all computers that share an AD DS domain.
waitfor	Sends or waits on a single.
wecutil	Creates and managements subscriptions for events.
winrm	Manages secure connections between local and remote computers via the WS management protocol.
winrs	Enables access to the command line of a remote computer via a secure connection to implement changes.
wsmanhttpconfig	Manages functions of the Windows Remote Management (winrm).
'''



"""
    domonic.d3.format
    ==================================

"""

from domonic.javascript import Math, Array, Number, String, Global, RegExp


def formatDecimal(x, ignore=None):
    x = Math.round(x)
    if Math.abs(x) >= 1e21:
        return x.toLocaleString("en").replace(r"/,/g", "")
    else:
        return Number(x).toString(10)


def formatDecimalParts(x, p=None):
    """[ Computes the decimal coefficient and exponent of the specified number x with
        significant digits p, where x is positive and p is in [1, 21] or undefined.
        For example, formatDecimalParts(1.23) returns ["123", 0].]

    Args:
        x ([type]): [description]
        p ([type]): [description]

    Returns:
        [type]: [description]
    """
    x = Number(x).toExponential() if p is None else Number(x).toExponential(p - 1)
    x2 = String(x).indexOf("e")
    i = x2
    if i < 0:
        return None  # NaN, ±Infinity

    coefficient = String(x).slice(0, i)

    # The string returned by toExponential either has the form \d\.\d+e[-+]\d+
    # (e.g., 1.2e+3) or the form \de[-+]\d+ (e.g., 1e+3).
    return [
        coefficient[0] + String(coefficient).slice(2)
        if len(coefficient) > 1
        else coefficient,
        String(x).slice(i + 1),  # .lstrip('+')
    ]


prefixExponent = None


def formatPrefixAuto(x, p):

    # print("formatPrefixAuto1",x,p)

    d = formatDecimalParts(x, p)
    if not d:
        return str(x)

    # print("formatPrefixAuto22",x,p)

    coefficient = d[0]
    if d[1] == "":
        d[1] = 0

    exponent = int(d[1])

    global prefixExponent
    prefixExponent = Math.max(-8, Math.min(8, Math.floor(exponent / 3))) * 3

    i = exponent - prefixExponent + 1
    n = len(coefficient)

    if i == n:
        return coefficient
    else:
        if i > n:
            return coefficient + "0".join(Array(i - n + 1))
        else:
            if i > 0:
                return (
                    String(coefficient).slice(0, i) + "." + String(coefficient).slice(i)
                )
            else:
                return (
                    "0."
                    + str("0".join(Array(1 - i)))
                    + formatDecimalParts(x, Math.max(0, p + i - 1))[0]
                )


def formatRounded(x, p):
    d = formatDecimalParts(x, p)
    if not d:
        return str(x)
    coefficient = d[0]

    if d[1] == "":
        d[1] = 0

    exponent = int(d[1])

    if exponent < 0:
        return "0." + Array(-exponent).join("0") + coefficient
    else:
        if len(coefficient) > (exponent + 1):
            return (
                String(coefficient).slice(0, exponent + 1)
                + "."
                + String(coefficient).slice(exponent + 1)
            )
        else:
            return coefficient + "0".join(Array(exponent - len(coefficient) + 2))


formatTypes = {
    "%": lambda x, p: Number(x * 100).toFixed(p),
    "b": lambda x, ignore=None: Number(Math.round(x)).toString(2),
    "c": lambda x, ignore=None: str(x),
    "d": formatDecimal,
    "e": lambda x, p: Number(x).toExponential(p),
    "f": lambda x, p: Number(x).toFixed(p),
    "g": lambda x, p: Number(x).toPrecision(p),
    "o": lambda x, ignore: Number(Math.round(x)).toString(8),
    "p": lambda x, p: formatRounded(x * 100, p),
    "r": formatRounded,
    "s": formatPrefixAuto,
    "X": lambda x, ignore=None: Number(Math.round(x)).toString(16).toUpperCase(),
    "x": lambda x, ignore=None: Number(Math.round(x)).toString(16),
}


def exponent(x):
    x = formatDecimalParts(Math.abs(x))
    return x[1] if x else None


def formatGroup(grouping, thousands):
    def func(value, width):

        i = len(str(value))
        t = []
        j = 0
        g = grouping[0]
        length = 0

        while i > 0 and g > 0:
            if (length + g + 1) > width:
                g = Math.max(1, width - length)

            a = i - g
            i -= g
            t.append(String(value).substring(a, i + g))

            length += g + 1
            if length > width:
                break

            j = (j + 1) % len(grouping)
            g = grouping[j]

        return thousands.join(Array(t).reverse())

    return func


def formatNumerals(numerals):
    def func(value):
        return value.replace(r"[0-9]g", lambda i: numerals[i])

    return func


re = r"^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$"


def formatSpecifier(specifier):
    match = RegExp(re).exec(str(specifier))
    if not match:
        raise Exception("invalid format: " + specifier)

    return FormatSpecifier(
        {
            "fill": match[0],
            "align": match[1],
            "sign": match[2],
            "symbol": match[3],
            "zero": match[4],
            "width": match[5],
            "comma": match[6],
            "precision": match[7] and String(match[7]).slice(1),
            "trim": match[8],
            "type": match[9],
        }
    )


class FormatSpecifier:
    def __init__(self, specifier):
        self.fill = (
            " " if specifier.get("fill", None) == None else str(specifier.get("fill"))
        )
        self.align = (
            ">" if specifier.get("align", None) == None else str(specifier.get("align"))
        )
        self.sign = (
            "-" if specifier.get("sign", None) == None else str(specifier.get("sign"))
        )
        self.symbol = (
            ""
            if specifier.get("symbol", None) == None
            else str(specifier.get("symbol"))
        )
        self.zero = bool(specifier.get("zero", False))
        self.width = (
            None if specifier.get("width", None) == None else specifier.get("width")
        )
        self.comma = bool(specifier.get("comma", None))
        self.precision = (
            None
            if specifier.get("precision", None) == None
            else specifier.get("precision")
        )
        self.trim = bool(specifier.get("trim", None))
        self.type = (
            "" if specifier.get("type", None) == None else str(specifier.get("type"))
        )

    def toString(self):
        z = "0" if self.zero else ""
        w = "" if self.width == None else Math.max(1, int(self.width) | 0)
        c = "," if self.comma else ""
        p = (
            ""
            if self.precision == None
            else "." + str(Math.max(0, int(self.precision) | 0))
        )
        t = "~" if self.trim else ""
        return (
            self.fill
            + self.align
            + self.sign
            + self.symbol
            + str(z)
            + str(w)
            + str(c)
            + str(p)
            + str(t)
            + self.type
        )

    def __str__(self):
        return self.toString()


# import formatTrim from "./formatTrim.js";

# // Trims insignificant zeros, e.g., replaces 1.2000k with 1.2k.
def formatTrim(s):
    #   out:
    #   for (var n = s.length, i = 1, i0 = -1, i1; i < n; ++i) {
    s = str(s)

    n = len(s)
    i = 1
    i0 = -1
    i1 = None

    #   for (var n = s.length, i = 1, i0 = -1, i1; i < n; ++i) {
    for i in range(1, n):
        if s[i] == ".":
            i0 = i1 = i
        elif s[i] == "0":
            if i0 == 0:
                i0 = i
                i1 = i
        # else:
        if not s[i]:
            break  # out
        if i0 > 0:
            i0 = 0

    return String(s).slice(0, i0) + String(s).slice(i1 + 1) if i0 > 0 else s


def identity(x):
    return x


map = Array.map
prefixes = [
    "y",
    "z",
    "a",
    "f",
    "p",
    "n",
    "µ",
    "m",
    "",
    "k",
    "M",
    "G",
    "T",
    "P",
    "E",
    "Z",
    "Y",
]


class NewFormatProxy:
    def __init__(self, nf, specifier):
        self.func = nf
        self.specifier = specifier

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __str__(self, *args, **kwargs):
        return str(self.specifier)


class formatLocale:
    def __init__(self, locale):

        self.group = (
            locale.get("grouping", None) == None or identity
            if locale.get("thousands", None) == None
            else formatGroup(
                [Global.Number(g) for g in locale["grouping"]],
                str(locale.get("thousands")),
            )
        )
        self.currencyPrefix = (
            ""
            if locale.get("currency", None) == None
            else str(locale.get("currency")[0])
        )
        self.currencySuffix = (
            "" if locale.get("currency") == None else str(locale.get("currency")[1])
        )
        self.decimal = (
            "." if locale.get("decimal", None) == None else str(locale.get("decimal"))
        )
        self.numerals = (
            identity
            if locale.get("numerals", None) == None
            else formatNumerals([str(n) for n in locale["numerals"]])
        )
        self.percent = (
            "%" if locale.get("percent", None) == None else str(locale["percent"])
        )
        self.minus = "−" if locale.get("minus", None) == None else str(locale["minus"])
        self.nan = "NaN" if locale.get("nan", None) == None else str(locale["nan"])

    def newFormat(self, specifier):
        specifier = formatSpecifier(specifier)

        fill = specifier.fill
        align = specifier.align
        sign = specifier.sign
        symbol = specifier.symbol
        zero = specifier.zero
        width = specifier.width
        comma = specifier.comma
        precision = specifier.precision
        trim = specifier.trim
        type = specifier.type

        if type == "n":
            comma = True
            type = "g"

        # The "" type, and any invalid type, is an alias for ".12~g".
        elif formatTypes.get(type, "") == "":
            if precision == None:
                precision = 12
                trim = True
                type = "g"

        # If zero fill is specified, padding goes after sign and before digits.
        if zero or (fill == "0" and align == "="):
            zero = True
            fill = "0"
            align = "="

        # Compute the prefix and suffix.
        # For SI-prefix, the suffix is lazily computed.
        prefix = ""
        if symbol == "$":
            prefix = self.currencyPrefix
        else:
            if symbol == "#" and RegExp(r"[boxX]").test(type):
                prefix = "0" + String(type).toLowerCase()
            else:
                prefix = ""

        suffix = ""
        if symbol == "$":
            suffix = self.currencySuffix
        else:
            suffix = self.percent if RegExp(r"[%p]").test(type) else ""

        # What format function should we use?
        # Is this an integer type?
        # Can this type generate exponential notation?
        formatType = formatTypes[type]
        maybeSuffix = RegExp(r"[defgprs%]").test(type)

        # Set the default precision if not specified,
        # or clamp the specified precision to the supported range.
        # For significant precision, it must be in [1, 21].
        # For fixed precision, it must be in [0, 20].
        if precision == None:
            precision = 6
            # precision = 0
        else:
            precision = (
                Math.max(1, Math.min(21, precision))
                if RegExp(r"[gprs]").test(type)
                else Math.max(0, Math.min(20, precision))
            )

        def format(value):
            valuePrefix = prefix
            valueSuffix = suffix
            # i = None
            # n = None
            # c = None

            if type == "c":
                valueSuffix = formatType(value) + valueSuffix
                value = ""
            else:
                value = value

                # Determine the sign. -0 is not less than 0, but 1 / -0 is!
                try:
                    valueNegative = value < 0
                except ZeroDivisionError:
                    valueNegative = 1 / value < 0
                except TypeError:
                    valueNegative = False

                # Perform the initial formatting.
                value = (
                    self.nan
                    if Global.isNaN(value)
                    else formatType(Math.abs(value), precision)
                )

                # Trim insignificant zeros.
                if trim:
                    value = formatTrim(value)

                # If a negative value rounds to zero after formatting,
                # and no explicit positive sign is requested, hide the sign.
                if valueNegative and value == 0 and sign != "+":
                    valueNegative = False

                # Compute the prefix and suffix.
                valuePrefix = (
                    (sign if sign == "(" else self.minus)
                    if valueNegative
                    else ""
                    if (sign == "-" or sign == "(")
                    else sign
                ) + valuePrefix

                global prefixExponent
                p1 = prefixes[int(8 + prefixExponent / 3)] if type == "s" else ""
                p2 = ")" if valueNegative and sign == "(" else ""
                valueSuffix = p1 + valueSuffix + p2

                # Break the formatted value into the integer “value” part that can be
                # grouped, and fractional or exponential “suffix” part that is not.
                if maybeSuffix:
                    i = -1
                    n = len(str(value))

                    while i < n:
                        c = String(value).charCodeAt(i)
                        if 48 > c or c > 57:
                            suff = (
                                self.decimal + String(value).slice(i + 1)
                                if c == 46
                                else String(value).slice(i)
                            )
                            valueSuffix = suff + valueSuffix
                            value = String(value).slice(0, i)
                            break
                        i += 1

            # If the fill character is not "0", grouping is applied before padding.
            if comma and not zero:
                value = self.group(value, Global.Infinity)

            # Compute the padding.
            length = len(valuePrefix) + len(str(value)) + len(valueSuffix)

            nonlocal width
            if width == "" or width == None:
                width = 0
            width = int(width)
            padding = (
                fill.join(Array(width - length + 1)) if length < int(width) else ""
            )

            # If the fill character is "0", grouping is applied after padding.
            if comma and zero:
                value = self.group(
                    padding + str(value),
                    width - len(valueSuffix) if len(padding) else Global.Infinity,
                )
                padding = ""

            # Reconstruct the final output based on the desired alignment.
            if align == "<":
                value = valuePrefix + value + valueSuffix + padding
            elif align == "=":
                value = valuePrefix + padding + str(value) + valueSuffix
            elif align == "^":
                length = len(padding) >> 1
                value = (
                    String(padding).slice(0, length)
                    + valuePrefix
                    + value
                    + valueSuffix
                    + String(padding).slice(length)
                )
            else:
                value = padding + valuePrefix + str(value) + valueSuffix

            return self.numerals(value)

        return NewFormatProxy(format, specifier)  # = newFormat

    def formatPrefix(self, specifier, value):
        specifier = formatSpecifier(specifier)
        specifier.type = "f"
        f = self.newFormat(specifier)
        e = Math.max(-8, Math.min(8, Math.floor(int(exponent(value)) / 3))) * 3
        k = Math.pow(10, -e)
        prefix = prefixes[int(8 + e / 3)]
        return lambda value: f(k * value) + prefix

    format = newFormat
    formatPrefix = formatPrefix


locale = None
format = None
formatPrefix = None


def defaultLocale(definition):
    global locale
    global format
    global formatPrefix
    locale = formatLocale(definition)
    format = locale.format
    formatPrefix = locale.formatPrefix  # ['formatPrefix']
    return locale


defaultLocale({"thousands": ",", "grouping": [3], "currency": ["$", ""]})


# https://github.com/d3/d3-format/tree/main/locale - TODO - finish all
def set_locale(code):
    """[sets the locale of the formatting engine]

    Args:
        code ([str]): [A language/country code i.e. en-GB, en-IN, en-US]

    Returns:
        [type]: [a new format obj]
    """
    loc = {
        "en-GB": {
            "decimal": ".",
            "thousands": ",",
            "grouping": [3],
            "currency": ["£", ""],
        },
        "en-IN": {
            "decimal": ".",
            "thousands": ",",
            "grouping": [3, 2, 2, 2, 2, 2, 2, 2, 2, 2],
            "currency": ["₹", ""],
        },
        "en-US": {
            "decimal": ".",
            "thousands": ",",
            "grouping": [3],
            "currency": ["$", ""],
        },
        "uk-UA": {
            "decimal": ",",
            "thousands": "\u00a0",
            "grouping": [3],
            "currency": ["", "\u00a0₴."],
        },
        "zh-CN": {
            "decimal": ".",
            "thousands": ",",
            "grouping": [3],
            "currency": ["¥", ""],
        },
        "sv-SE": {
            "decimal": ",",
            "thousands": "\u00a0",
            "grouping": [3],
            "currency": ["", " kr"],
        },
        "sl-SI": {
            "decimal": ",",
            "thousands": ".",
            "grouping": [3],
            "currency": ["", "\u00a0€"],
        },
        "ru-RU": {
            "decimal": ",",
            "thousands": "\u00a0",
            "grouping": [3],
            "currency": ["", "\u00a0руб."],
        },
        "pt-PT": {
            "decimal": ",",
            "thousands": "\u00a0",
            "grouping": [3],
            "currency": ["", "\u00a0€"],
        },
        "pt-BR": {
            "decimal": ",",
            "thousands": ".",
            "grouping": [3],
            "currency": ["R$", ""],
        },
        "pl-PL": {
            "decimal": ",",
            "thousands": ".",
            "grouping": [3],
            "currency": ["", "zł"],
        },
        "nl-NL": {
            "decimal": ",",
            "thousands": ".",
            "grouping": [3],
            "currency": ["€\u00a0", ""],
        },
        "mk-MK.": {
            "decimal": ",",
            "thousands": ".",
            "grouping": [3],
            "currency": ["", "\u00a0ден."],
        },
        "ko-KR": {
            "decimal": ".",
            "thousands": ",",
            "grouping": [3],
            "currency": ["₩", ""],
        },
        "ja-JP": {
            "decimal": ".",
            "thousands": ",",
            "grouping": [3],
            "currency": ["", "円"],
        },
        "it-IT": {
            "decimal": ",",
            "thousands": ".",
            "grouping": [3],
            "currency": ["€", ""],
        },
        "hu-HU.": {
            "decimal": ",",
            "thousands": "\u00a0",
            "grouping": [3],
            "currency": ["", "\u00a0Ft"],
        },
        "he-IL": {
            "decimal": ".",
            "thousands": ",",
            "grouping": [3],
            "currency": ["₪", ""],
        },
        "fr-FR": {
            "decimal": ",",
            "thousands": "\u00a0",
            "grouping": [3],
            "currency": ["", "\u00a0€"],
            "percent": "\u202f%",
        },
        "es-ES": {
            "decimal": ",",
            "thousands": ".",
            "grouping": [3],
            "currency": ["", "\u00a0€"],
        },
        "de-DE": {
            "decimal": ",",
            "thousands": ".",
            "grouping": [3],
            "currency": ["", "\u00a0€"],
        },
        "ar-YE": {
            "decimal": "\u066b",
            "thousands": "\u066c",
            "grouping": [3],
            "currency": ["", " \u0631\u002e\u0649\u002e"],
            "numerals": [
                "\u0660",
                "\u0661",
                "\u0662",
                "\u0663",
                "\u0664",
                "\u0665",
                "\u0666",
                "\u0667",
                "\u0668",
                "\u0669",
            ],
        },
    }[code]
    return formatLocale(loc)


# formatDefaultLocale
# format
# precisionFixed
# precisionPrefix
# precisionRound



"""
    domonic.webapi.geolocation
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Geolocation_API
"""

from domonic.events import Event
from domonic.javascript import Error


class Geolocation:

    @staticmethod
    def getCurrentPosition(successCallback, errorCallback, options):
        pass

    @staticmethod
    def watchPosition(successCallback, errorCallback, options):
        pass

    @staticmethod
    def clearWatch(watchId):
        pass


class GeolocationPosition():
    coords = None
    timestamp = None


class GeolocationCoordinates():
    latitude = None
    longitude = None
    altitude = None
    accuracy = None
    altitudeAccuracy = None
    heading = None
    speed = None


class GeolocationError(Error):

    PERMISSION_DENIED = 1
    POSITION_UNAVAILABLE = 2
    TIMEOUT = 3

    # def __init__(self, code, message,):
    #     self.code = code
        # self.message = message
        # super().__init__(message)



"""
    domonic.webapi.urlpattern
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/URLPattern
"""


class URLPattern:

    def __init__(self, pattern):
        """[matches URLs or parts of URLs against a pattern.
        The pattern can contain capturing groups that extract parts of the matched URL.]

        Args:
            pattern (str): [The pattern to match against.]
        """
        self.pattern = pattern
        self.hash = ''
        self.hostname = ''
        self.href = ''
        self.origin = ''
        self.pathname = ''
        self.port = ''
        self.protocol = ''
        self.search = ''
        self.username = ''

    def __str__(self):
        return self.pattern

    # https://developer.mozilla.org/en-US/docs/Web/API/URLPattern/exec #TODO - unit tests from this page
    def exec_(self, input, baseURL=None):
        """[Returns an object with the matched parts of the URL or None if the URL does not match.]

        Args:
            url (str): [The URL or URL parts to match against]
            baseURL (str, optional): [the base URL to use in cases where input is a relative URL]. Defaults to None.
        """
        # if baseURL is None:
        #     baseURL = ''
        # else:
        #     baseURL = baseURL.href
        # if input == self.pattern:
        #     return self.href
        # else:
        #     return None
        raise NotImplementedError

    # TODO - unit tests 
    def test(self, url):
        """[summary]

        Args:
            url (str): [description]
        """
        raise NotImplementedError


"""
    domonic.game
    ====================================

"""
import random


class Game():

    @staticmethod
    def roll_dice(sides=6) -> int:
        """[rolls a dice]

        Args:
            sides (int, optional): [number of sides on the dice]. Defaults to 6.

        Returns:
            [int]: [number rolled]
        """
        return random.randint(1, sides)

    @staticmethod
    def pick_a_card() -> str:
        """[selects a random suit and card]

        Returns:
            [str]: [a card from the deck in the form of a string. i.e '♥A']
        """
        suits = ["♠", "♥", "♦", "♣"]
        cards = ["2", "3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K", "A"]
        return random.choice(cards) + random.choice(suits)

    @staticmethod
    def deal_cards(n: int=1) -> list:
        """[deals n cards]

        Args:
            n (int, optional): [number of cards to deal]. Defaults to 1.

        Returns:
            [list]: [list of cards]
        """
        return [Game.pick_a_card() for _ in range(n)]  # TODO - this could have duplicates

    @staticmethod
    def random_bool() -> bool:
        #  (https://6b.eleuther.ai/) TODO - test
        rv = random.Random()
        if rv.uniform(0.0, 1.0):
            return True
        else:
            return False

    @staticmethod
    def random_char() -> str:
        return chr(random.choice(range(ord('A'), ord('Z') + 1)))

    # @staticmethod
    # def bingo():
        # calls=["Kelly's Eye","One Little Duck","Cup of Tea","Knock at the Door","Man Alive","Half a Dozen","Lucky Seven","Garden Gate","Doctor's Order","Prime Minister's Den","Legs Eleven","One Dozen","Unlucky for Some","Valentine's Day","Young and Keen","Sweet Sixteen and Never Been Kissed","Dancing Queen","Coming of Age","Goodbye Teens","One Score","Key of the Door","Two Little Ducks","Thee and Me","Two Dozen","Duck and Dive","Pick and Mix","Gateway to Heaven","In a State","Rise and Shine","Dirty Gertie","Get up and Run","Buckle my Shoe","Fish, Chips and Peas","Ask for More","Jump and Jive","Three Dozen","More than Eleven","Christmas Cake","39 Steps","Life Begins","Time for Fun","Winnie-the-Pooh","Down on your Knees","Droopy Drawers","Halfway There","Up to Tricks","Four and Seven","Four Dozen","PC","Half a Centry","Tweak of the Thumb","Danny La Rue","Stuck in a Tree","Clean the Floor","Snakes Alive","Shotts Bus","Heinz Varieties","Make them Wait","Brighton Line","Five Dozen","Baker's Bun","Tickety-Boo","Tickle Me Sixty Three","Red Raw","Old Age Pension","Clickety Click","Stairway to Heaven","Saving Grace","Favourite of Mine","Three Score and Ten","Bang on the Drum","Six Dozen","Queen Bee","Hit the Floor","Strive and Strive","Trombones","Sunset Strip","39 More Steps","One More Time","Eight and Blank","Stop and Run","Straight on Through","Time for Tea","Seven Dozen","Staying Alive","Between the Sticks","Torquay in Devon","Two Fat Ladies","Nearly There","Top of the Shop"]
        # index = Math.random()*len(calls)
        # return calls[index], index+1

    # @staticmethod
    # def random_name():

    # @staticmethod
    # def random_address():



"""
    test_templates
    ~~~~~~~~~~~~
    tests for templates
"""

import unittest
# import json
# import requests
# from mock import patch

from domonic.html import *
from domonic.templates import status_page

class TestCase(unittest.TestCase):

    def test_404(self):
        p = status_page(404)
        print(p)
        # self.assertEqual(p.status, 404)

    def test_401(self):
        # BUG - notice it tabs in due to being parent of <html>
        p = status_page(401, False)
        print(p)
        # self.assertEqual(p.status, 404)




if __name__ == '__main__':
    unittest.main()



"""
    domonic.ext
    ====================================

    This dir is for extending domonic to be useable with other python libs.

"""

# HELLO WORLDS - hello world code for other libs

HELLO_SANIC: str = """
from sanic import Sanic
from sanic import response
from domonic.html import *

app = Sanic("My Hello, world app")

@app.route('/')
async def test(request):
    return response.html(str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        ))
    )

if __name__ == '__main__':
    app.run()
"""

HELLO_FLASK: str = """
from flask import Flask
from domonic.html import *

app = Flask(__name__)

@app.route("/")
def hello():
    return str(
        html(
        head(),
        body(
            div(span("Hello World!"))
            )
        ))

if __name__ == '__main__':
    app.run()
"""

HELLO_CHERRYPY: str = """
import cherrypy
from domonic.html import *

class HelloWorld(object):

    @cherrypy.expose
    def index(self):
        return str(
                    html(
                    head(),
                    body(
                        div(span("Hello, World!"))
                        )
                    )
                )

cherrypy.quickstart(HelloWorld())
"""


HELLO_BOTTLE: str = """
from bottle import route, run
from domonic.html import *

@route('/hello/<name>')
def index(name):
    return str(
            html(
            head(),
            body(
                div(span(f"Hello, {name}!"))
                )
            )
        )

run(host='localhost', port=8080)
"""

HELLO_AIOHTTP: str = """
from domonic.html import *
from aiohttp import web

async def handle(request):
    name = request.match_info.get('name', "Anonymous")
    page = html(head(),body(div(span("Hello, World!"))))
    return web.Response(text=str(page), content_type='text/html')

app = web.Application()
app.add_routes([web.get('/', handle),
                web.get('/{name}', handle)])

if __name__ == '__main__':
    web.run_app(app)
"""

HELLO_TORNADO: str = """
import tornado.ioloop
import tornado.web
from domonic.html import *

class MainHandler(tornado.web.RequestHandler):
    def get(self):
        self.write(str(
            html(
            head(),
            body(
                div(span("Hello World!"))
                )
            )))

def make_app():
    return tornado.web.Application([
        (r"/", MainHandler),
    ])

if __name__ == "__main__":
    app = make_app()
    app.listen(8888)
    tornado.ioloop.IOLoop.current().start()
"""

HELLO_WERKZEUG: str = """
from werkzeug.wrappers import Request, Response
from domonic.html import *

@Request.application
def application(request):
    return Response(str(
                    html(
                    head(),
                    body(
                        div(span("Hello World!"))
                        )
                    )), mimetype='text/html')

if __name__ == '__main__':
    from werkzeug.serving import run_simple
    run_simple('localhost', 4000, application)
"""


def get_hello_world(server):
    hello_words = {
        'flask': HELLO_FLASK,
        'cherrypy': HELLO_CHERRYPY,
        'sanic': HELLO_SANIC,
        'bottle': HELLO_BOTTLE,
        'aiohttp': HELLO_AIOHTTP,
        'tornado': HELLO_TORNADO,
        'werkzeug': HELLO_WERKZEUG
    }
    if server in hello_words:
        return hello_words[server]
    else:
        return None



"""
    domonic.webapi.netinfo
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Network_Information_API
"""

from domonic.dom import EventTarget

class NetworkInformation(EventTarget):

    def __init__(self):
        self.type = None
        self.downlinkMax = None
        self.effectiveType = None
        self.rtt = None
        self.saveData = None
        self.downlink = None
        self.effectiveType = None
        self.rtt = None

    def __str__(self):
        return 'type: {}, downlinkMax: {}, effectiveType: {}, rtt: {}, saveData: {}, downlink: {}, effectiveType: {}, rtt: {}'.format(
            self.type, self.downlinkMax, self.effectiveType, self.rtt, self.saveData, self.downlink, self.effectiveType, self.rtt
        )



"""
    test_svg
    ~~~~~~~~~~~~
    - unit tests for svg
"""

import unittest
# import requests
# from mock import patch

from domonic import domonic

from domonic.svg import *
# from domonic.decorators import silence


class TestCase(unittest.TestCase):

    # @silence
    def test_domonic_svg(self):
        mysvg = svg()
        assert str(mysvg) == '<svg></svg>'
        mysvg.appendChild(circle(_cx="50", _cy="50", _r="40", _stroke="green", **{"_stroke-width": "4"}, _fill="yellow"))
        assert str(mysvg) == '<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"></circle></svg>'
        mysvg.appendChild(circle(_cx="50", _cy="50", _r="40", _stroke="green", **{"_stroke-width": "4"}, _fill="yellow"))
        assert str(mysvg) == '<svg><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"></circle><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"></circle></svg>'
        # assert mysvg.toxml() == '<?xml version="1.0" encoding="utf-8"?>\n<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="100" height="100" viewBox="0 0 100 100">\n</svg>'

    def test_domonic_cirlce(self):
        test = svg(
            circle(_cx="50", _cy="50", _r="40", _stroke="green", **{"_stroke-width": "4"}, _fill="yellow"),
            _width="100", _height="100",
        )
        # print(test)
        assert str(test) == '<svg width="100" height="100"><circle cx="50" cy="50" r="40" stroke="green" stroke-width="4" fill="yellow"></circle></svg>'

    def test_domonic_node(self):
        circ = svg(
            circle(_cx="50", _cy="50", _r="40", _stroke="green", **{"_stroke-width": "4"}, _fill="yellow"),
            _width="100", _height="100",
        )
        mysvg = svg()
        mysvg.appendChild(circ / 10)
        # print(mysvg)

    # def test_hyphen_elements(self):
    #     test = svg(
    #         missing_glyph(),
    #     )
    #     print(test)

    # def test_font(self):
    #     test = svg(
    #         font_face(),
    #     )
    #     print(test)


if __name__ == '__main__':
    unittest.main()



"""
    test_domonic
    ~~~~~~~~~~~~
    - unit tests for domonic
    # TODO - tests for all bs5 pages
"""

import unittest
# import requests
# from mock import patch

# from domonic import domonic

from domonic.xml.mathml import *
# from domonic.decorators import silence


class TestCase(unittest.TestCase):

    def test_mathml(self):
        somemath = math_(
            maction('x'),
            menclose('x'),
            merror('x'),
            mfenced('x'),
            mfrac('x'),
            mi('x'),
            mmultiscripts('x'),
            mn('x'),
            mo('x'),
            mover('x'),
            mpadded('x'),
            mphantom('x'),
            mroot('x'),
            mrow('x'),
            ms('x'),
            mspace('x'),
            msqrt('x'),
            mstyle('x'),
            msub('x'),
            msubsup('x'),
            msup('x'),
            mtable('x'),
            mtd('x'),
            mtext('x'),
            mtr('x'),
            munder('x'),
            munderover('x'),
            semantics('x'),
            maligngroup('x'),
            malignmark('x'),
            msline('x'),
            msgroup('x'),
            mlongdiv('x'),
            mstyle('x'),
            mprescripts('x'),
            mscarries('x'),
            mscarry('x'),
            munder('x'),
            munderover('x'),
            none('x'),
        )

        print(somemath)


if __name__ == '__main__':
    unittest.main()



"""
    domonic.d3.tile
    ====================================

    # TODO - completely untested

"""

from domonic.javascript import Math


def defaultScale(t):
    return t.k


def defaultTranslate(t):
    return [t.x, t.y]


def constant(x):
    return lambda: x


class tile():

    def __init__(self, *args):
        self.x0 = 0
        self.y0 = 0
        self.x1 = 960
        self.y1 = 500
        self.clampX = True
        self.clampY = True
        self.tileSize = 256
        self.scale = defaultScale
        self.translate = defaultTranslate
        self.zoomDelta = 0

    def __call__(self, *args):
        scale_ = self.scale.apply(self, args)
        translate_ = self.translate.apply(self, args)
        z = Math.log2(scale_ / self.tileSize)
        z0 = Math.round(Math.max(z + self.zoomDelta, 0))
        k = Math.pow(2, z - z0) * self.tileSize
        x = translate_[0] - scale_ / 2
        y = translate_[1] - scale_ / 2
        xmin = Math.max(0 if self.clampX else -Infinity, Math.floor((self.x0 - x) / k))
        xmax = Math.min(1 << z0 if self.clampX else Infinity, Math.ceil((self.x1 - x) / k))
        ymin = Math.max(0 if self.clampY else -Infinity, Math.floor((self.y0 - y) / k))
        ymax = Math.min(1 << z0 if self.clampY else Infinity, Math.ceil((self.y1 - y) / k))
        tiles = []

        for y in range(ymin, ymax):
            for x in range(xmin, xmax):
                tiles.append([x, y, z0])

        tiles.translate = [x / k, y / k]
        tiles.scale = k
        return tiles

    def size(_, *args):
        return (x0 = y0 = 0, x1 = +_[0], y1 = +_[1], tile) if Global.Boolean(args) else [x1 - x0, y1 - y0]

    def extent(_, *args):
        return (x0 = +_[0][0], y0 = +_[0][1], x1 = +_[1][0], y1 = +_[1][1], tile) if Global.Boolean(args) else [[x0, y0], [x1, y1]]

    # def scale(_, *args):
    #     return Global.Boolean(args) ? (scale = typeof _ === "function" ? _ : constant(+_), tile) : scale

    # def translate(_, *args):
    #     return Global.Boolean(args) ? (translate = typeof _ === "function" ? _ : constant([+_[0], +_[1]]), tile) : translate

    def zoomDelta(_, *args):
        return (zoomDelta = +_, tile) if Global.Boolean(args) else zoomDelta

    def tileSize(_, *args):
        return (tileSize = +_, tile) if Global.Boolean(args) else tileSize

    # def clamp(_, *args):
    #     return (clampX = clampY = !!_, tile) if Global.Boolean(args) else clampX and clampY

    # def clampX(_, *args):
    #     return (clampX = !!_, tile) if Global.Boolean(args) else clampX

    # def clampY(_, *args):
    #     return (clampY = !!_, tile) if Global.Boolean(args) else clampY

#   return tile


"""
    domonic.i18n.mi
    ====================================

    html tags in Maori

    first translated to Maori using google translate.
    send a pull request if you have a better translation.
"""
from domonic.html import *

tinana = body
upoko = head
tuhi = script
kāhua = style
tomokanga = portal
kaha = strong
kupuporaka = blockquote
tepu = table
ahua = form
tapanga = label
tuku = submit
taitara = title
# noscript = noscript
wahanga = section
# nav = nav
tuhinga = article
tahae = aside
# hgroup = hgroup
wāhitau = address
ahua = figure
whakakitenga = figcaption
iti = small
whakahua = cite
waehere = code
iti = sub
sup = sup
tohu = mark
rupi = ruby
whanganga = span
iframe = iframe
ataata = video
ororongo = audio
kākahu = canvas
tapanga = caption
roopuroopu = colgroup
tinana = tbody
# upoko = thead
waewae = tfoot
huingamara = fieldset
pūrākau = legend
patene = button
whiriwhiri = select
rārangiraraunga = datalist
optgroup = optgroup
kōwhiringa = option
textarea = textarea
putanga = output
ahuwhakamua = progress
mita = meter
taipitopito = details
whakarāpopototanga = summary
tahua = menu
tahua = menuitem
momotuhi = font
pane = header
hiku = footer
turanga = base
hono = link
meta = meta
img = img
tāmau = embed
puna = source
ara = track
rohe = area
# col = col
tāuru = input
whakahau = command
matua = main
momotuhituranga = basefont
pokapū = center
tāmau = embed
momotuhi = font
isindex = isindex
whakarārangi = listing
Kuputuhi = plaintext
patuki = strike
tauira = template
pikitia = picture
momokorero = doctype
korero = comment



"""
    domonic.geom
    ====================================

    written by.ai

"""
import math
from domonic.javascript import Math

# from domonic.vector import vector
from domonic.geom.vec2 import vec2
from domonic.geom.vec3 import vec3
from domonic.geom.vec4 import vec4

# from domonic.mat4 import mat4
# from domonic.mat3 import mat3
# from domonic.quat import quat

# from domonic.color import color

# from domonic.point import point
# from domonic.matrix import matrix
# from domonic.transform import transform
# from domonic.matrix_utils import *
# from domonic.geom_utils import *
# from domonic.geom_math import *
# from domonic.geom_shader import *
# from domonic.geom_shader_utils import *
# from domonic.geom_shader_math import *
# from domonic.geom_shader_math_utils import *

from domonic.geom.shape import Shape
from domonic.geom.shape import Point
from domonic.geom.shape import Line
# from domonic.geom.shape import Plane
from domonic.geom.shape import Circle
from domonic.geom.shape import Polygon
# from domonic.geom.shape import Triangle
# from domonic.geom.shape import Quad
# from domonic.geom.shape import Hexagon
# from domonic.geom.shape import Rectangle
from domonic.geom.shape import Rect
# from domonic.geom.shape import Ellipse
# from domonic.geom.shape import BoundingBox
# from domonic.geom.shape import BoundingCircle


# __all__ # TODO - i think this is i need to use


class matrix(object):
    """[matrixs]
    """

    def __init__(self, m):
        self.m = m

    def translate(self, pt):
        """ Translates the point on the vector defined by the vectors m[0] and m[1]."""
        return vec2(self.m[0][0], self.m[0][1]) + pt - self.m[0][0]

    def rotate(self, pt):
        """ Rotates the point on the vector defined by the vectors m[0] and m[1]."""
        return vec2(self.m[1][0], self.m[1][1]) + pt - self.m[1][0]

    def scale(self, pt):
        """ Scales the point on the vector defined by the vectors m[0] and m[1]."""
        return vec2(self.m[2][0], self.m[2][1]) + pt - self.m[2][0]


class Quaternion():

    def __init__(self, w, x, y, z):
        self.w = w
        self.x = x
        self.y = y
        self.z = z
        self.q = [w, x, y, z]


class Plotter(object):
    def __init__(self, canvas=None):
        self.canvas = canvas  # or canvas.Canvas('Example Plotter', 500, 500)
        self.points = []
        self.starting_color = (255, 255, 255)
        self.color = self.starting_color
        self.clear()

    def add_point(self, point):
        self.points.append(point)

    # def clear(self):
    #     pass

    # def plot(self):
    #     for p in self.points:
    #         self.canvas.draw_point(p)


class Path(object):
    def __init__(self, points=None):
        self.points = points or []

    def add_point(self, point):
        self.points.append(point)

    # def get_points(self):
    #     return self.points

    # def get_points_as_vector(self):
    #     return vec3(self.points)

    # def get_points_as_matrix(self):
    #     return mat3(self.points)


# class Cursor(object):


class Group(object):
    def __init__(self, shapes=None):
        self.shapes = shapes or []

    def add_shape(self, shape):
        self.shapes.append(shape)

    def get_shapes(self):
        return self.shapes


class Layer(object):

    def __init__(self, name=None):
        self.name = name
        self.shapes = []
        self.visible = True
        self.selected = False
        self.parent = None
        self.children = []
        self.parent = None
        self.zindex = 0
        self.color = (0, 0, 0)
        self.alpha = 1.0

    def add(self, shape):
        self.shapes.append(shape)
        return self

    def add_at(self, shape, index):
        self.shapes.insert(index, shape)
        return self

    def has(self, shape):
        return shape in self.shapes

    def get_at(self, index):
        return self.shapes[index]

    def get_by_name(self, name):
        for shape in self.shapes:
            if shape.name == name:
                return shape
        return None

    def get_by_property(self, property):
        return [shape for shape in self.shapes if shape.has_property(property)]

    def remove(self, shape):
        self.shapes.remove(shape)
        return self

    def remove_at(self, index):
        self.shapes.pop(index)
        return self

    def swap(self, shape1, shape2):
        self.shapes[self.shapes.index(shape1)] = shape2

    def swap_at(self, index1, index2):
        self.shapes[index1], self.shapes[index2] = self.shapes[index2], self.shapes[index1]
        return self

    def show(self):
        for shape in self.shapes:
            shape.show()
        return self

    def hide(self):  # hide all shapes
        for shape in self.shapes:
            shape.hide()
        return self

    def delete(self):
        for shape in self.shapes:
            shape.delete()
        return self

    def __len__(self):
        return len(self.shapes)

    def __iter__(self):
        return iter(self.shapes)

    def __getitem__(self, key):
        return self.shapes[key]

    def __setitem__(self, key, value):
        self.shapes[key] = value

    def __add__(self, other):
        self.shapes += other.shapes
        return self

    def __sub__(self, other):
        self.shapes -= other.shapes
        return self

    def __mul__(self, other):
        self.shapes *= other
        return self

    def __rmul__(self, other):
        self.shapes *= other
        return self

    def __iadd__(self, other):
        self.shapes += other.shapes
        return self

    def __isub__(self, other):
        self.shapes -= other.shapes
        return self

    def __imul__(self, other):
        self.shapes *= other
        return self

    def __idiv__(self, other):
        self.shapes /= other
        return self

    def __itruediv__(self, other):
        self.shapes /= other
        return self

    def __imod__(self, other):
        self.shapes %= other
        return self

    def __ipow__(self, other):
        self.shapes **= other
        return self

    def __ilshift__(self, other):
        self.shapes <<= other
        return self

    def __irshift__(self, other):
        self.shapes >>= other
        return self

    def __iand__(self, other):
        self.shapes &= other
        return self


class Timeline(object):

    def __init__(self):
        self.layers = []
        # self.current_layer = None
        self.current_frame = 0
        self.enabled = True
        self.is_playing = False
        self.play_speed = 1.0
        self.total_frames = 0
        self.framerate = 0

    def go_to_frame(self, frame):
        self.current_frame = frame

    def next_frame(self):
        self.current_frame += 1

    def prev_frame(self):
        self.current_frame -= 1

    def stop(self):
        self.is_playing = False

# class Interactive(object):#??
    # def handle_click_event(self):

    # onFrame
    # onMouseDown
    # onMouseDrag
    # onMouseUp
    # onClick
    # onDoubleClick
    # onMouseMove
    # onMouseEnter
    # onMouseLeave



"""
    domonic.webapi.mediacapabilities
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Media_Capabilities_API
"""

class MediaCapabilities:
    
    @staticmethod
    def encodingInfo(mediaEncodingConfiguration):
        pass

    @staticmethod
    def decodingInfo(mediaDecodingConfiguration):
        pass


# MediaDecodingConfiguration
# MediaEncodingConfiguration
# VideoConfiguration
# AudioConfiguration



"""
    domonic.sitemap
    ====================================

    generate or load sitemaps

    warning - when using image and video tags from this package they will be namespaced i.e <image:image> and <video:video>
    so i'd advise to only import them within the def that you use them in to avoid conflict with html.image

"""

import datetime

from domonic.html import tag  # , closed_tag
from domonic.dom import Element, Document

# __all__ = ['sitemap', 'url', 'lastmod']

sitemap_tags = ["sitemapindex", "sitemap", "urlset", "url", "loc", "lastmod", "changefreq", "priority", 
                "image:image", "image:loc"]

sitemap_attributes = ["xmlns", "xmlns:xsi", "xsi:schemaLocation",
                      "xmlns:xhtml", "xmlns:xlink", "xmlns:atom", "xmlns:geo"]
# sitemap_change_frequencies = ['always', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'never']

XMLNS = 'http://www.sitemaps.org/schemas/sitemap/0.9'
XMLNS_XSI = 'http://www.w3.org/2001/XMLSchema-instance'
SCHEMA_SITEINDEX = 'http://www.sitemaps.org/schemas/sitemap/0.9/siteindex.xsd'
SCHEMA_SITEMAP = 'http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd'
#     xmlns="http://www.google.com/schemas/sitemap-image/1.1"
#     xmlns:video="http://www.google.com/schemas/sitemap-video/1.1">


def sitemap_init(self, *args, **kwargs):
    tag.__init__(self, *args, **kwargs)
    Element.__init__(self, *args, **kwargs)


def sitemap_format(self, *args, **kwargs):
    """ attempts to prettify the output of the sitemap. """
    outp = f"<{self.name}{self.__attributes__}>{self.content}</{self.name}>"
    from domonic.xml import prettify
    return prettify(outp)


sitemapindex = type('sitemapindex', (tag, Document), {
    'name': 'sitemapindex',
    'xmlns': XMLNS,
    'xmlns:xsi': XMLNS_XSI,
    'xsi:schemaLocation': SCHEMA_SITEINDEX,
    '__init__': sitemap_init,
    '__format__': sitemap_format})

sitemap = type('sitemap', (tag, Element), {'name': 'sitemap', '__init__': sitemap_init})

urlset = type('urlset', (tag, Element), {
    'name': 'urlset',
    'xmlns:xsi': XMLNS_XSI,
    'xsi:schemaLocation': SCHEMA_SITEMAP,
    'xmlns': XMLNS,
    '__init__': sitemap_init})


def url_init(self, *args, **kwargs):
    # validate kwargs
    # if 'xmlns' in kwargs:
    #     if kwargs['xmlns'] not in sitemap_attributes:
    #         raise ValueError("xmlns attribute not valid")
    # if 'lastmod' in kwargs:
    #     if not isinstance(kwargs['lastmod'], datetime.datetime):
    #         raise ValueError("lastmod must be a datetime object")
    # else:
    #     kwargs['lastmod'] = datetime.datetime.now()

    # if lastmod is a datetime convert it to an lastmod object
    # if 'lastmod' in kwargs:
    #     if not isinstance(kwargs['lastmod'], datetime.datetime):
    #         kwargs['lastmod'] = lastmod(datetime.datetime.now())
    #     else:
    #         kwargs['lastmod'] = lastmod(kwargs['lastmod'])

    # if 'priority' in kwargs:
    #     if kwargs['priority'] not in range(0, 1):
    #         raise ValueError("priority must be between 0 and 1")

    # if 'changefreq' in kwargs:
    #     if kwargs['changefreq'] not in ('always', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'never'):
    #         raise ValueError("changefreq must be one of: always, hourly, daily, weekly, monthly, yearly, never")

    # if 'loc' in kwargs:
    #     if not isinstance(kwargs['loc'], str):
    #         raise ValueError("loc must be a string")
        # escape the loc string ?
        # kwargs['loc'] = kwargs['loc'].replace('&', '&amp;')
        # kwargs['loc'] = kwargs['loc'].replace('<', '&lt;')
        # kwargs['loc'] = kwargs['loc'].replace('>', '&gt;')
        # kwargs['loc'] = kwargs['loc'].replace('"', '&quot;')
        # kwargs['loc'] = kwargs['loc'].replace("'", '&apos;')

    tag.__init__(self, *args, **kwargs)
    Element.__init__(self, *args, **kwargs)


url = type('url', (tag, Element), {
    'name': 'url',
    # loc = loc
    # lastmod = lastmod
    # changefreq = changefreq
    # priority = priority
    '__init__': sitemap_init})

loc = type('loc', (tag, Element), {'name': 'loc', '__init__': sitemap_init})
lastmod = type('lastmod', (tag, Element), {'name': 'lastmod', '__init__': sitemap_init})
changefreq = type('changefreq', (tag, Element), {'name': 'changefreq', '__init__': sitemap_init})
priority = type('priority', (tag, Element), {'name': 'priority', '__init__': sitemap_init})


def sitemapindex_from_urls(urls):
    """
    Create a sitemap index from a list of urls.

    WARNING:
    there's a difference between a sitemap index and a sitemap. 
    make sure you know what you want.

    # i.e

    # <?xml version="1.0" encoding="UTF-8"?>
    # <sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
    # 	<sitemap>
    # 		<loc>https://xyz.com/sitemap1.xml</loc>
    # 		<lastmod>2021-07-08T13:12:16+00:00</lastmod>
    # 	</sitemap>
    # </sitemapindex>

    """
    sitemap_index = sitemapindex()
    for url in urls:
        sitemap_index.append(sitemap(loc(url), lastmod(datetime.datetime.now())))


def sitemap_from_urls(urls):
    """
    Create a sitemap from a list of urls.add()

    Note: This won't allow you to add priority or changefreq of the urls. or add images etc
    tho u could loop the nodes afterwards and do that.

    WARNING:
    there's a difference between a sitemap index and a sitemap. 
    make sure you know what you want.

    """
    sitemap = urlset()
    for url in urls:
        sitemap.append(url(loc(url), lastmod(datetime.datetime.now())))
    return sitemap


def get_sitemap(path: str, *args, **kwargs):
    # use requests to downlaod a sitemap
    import requests
    r = requests.get(path)
    # if r.status_code == 200:
    #     return r.text
    # else:
    #     raise ValueError("Could not get sitemap")

    import domonic
    some_sitemap = domonic.domonic.parseString(r.text)
    # print('RES:', some_sitemap)
    # print('RES:', type(some_sitemap))
    # print('RES:', str(some_sitemap))
    # print( some_sitemap.getElementByTagName('sitemap') )
    # return some_sitemap
    return some_sitemap


# def submit_sitemap(sitemap_url: str, search_engine_url: str):
#     """
#     submit a sitemap to a search engine.

#     """
#     import requests
#     r = requests.get(search_engine_url + '/ping?sitemap=' + sitemap_url)
#     print(r.text)



# a regular site map

# <?xml version="1.0" encoding="UTF-8"?>
# <?xml-stylesheet type="text/xsl" href="seo/css/main-sitemap.xsl"?>
# <urlset xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
# xmlns:image="http://www.google.com/schemas/sitemap-image/1.1" 
# xsi:schemaLocation="http://www.sitemaps.org/schemas/sitemap/0.9 http://www.sitemaps.org/schemas/sitemap/0.9/sitemap.xsd http://www.google.com/schemas/sitemap-image/1.1 http://www.google.com/schemas/sitemap-image/1.1/sitemap-image.xsd" 
# xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
	# <url>


# note on namespaces

# <image:image>
# <image:image>

# represent that image node as a python class?
# image(_namespace="http://www.google.com/schemas/sitemap-image/1.1", )

globals()['image:image'] = type('image:image', (tag, Element), {
    'name': 'image',
    'ns': 'image',
    '__init__': sitemap_init})

globals()['image:loc'] = type('image:loc', (tag, Element), {
    'name': 'loc',
    'ns': 'image',
    '__init__': sitemap_init})


#image tag example

# <urlset
#     xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"
#     xmlns="http://www.google.com/schemas/sitemap-image/1.1"
#     xmlns:video="http://www.google.com/schemas/sitemap-video/1.1">
    #     <image:image>
    #       <image:loc>http://e.com/img.png</image:loc>
    #       <image:caption>Stunning Men's Hats</image:caption>
    #       <image:geo_location>Boise, Idaho, USA</image:geo_location>
    #       <image:title>Men's Hat Image</image:title>
    #    </image:image>
# </urlset>



"""
    test_domonic
    ~~~~~~~~~~~~
    unit tests for css
"""

import unittest

import json
# import requests
# from mock import patch

# from domonic.html import *
from domonic.JSON import JSON, return_json


class TestCase(unittest.TestCase):

    SOMEJSON = '''
    { "items": [
        {
            "id":"01",
            "name": "Java",
            "thing":{},
            "list":[]
        },
        {
            "id":"07",
            "name": "C++",
            "thing":{},
            "list":[]
        }
    ]}
    '''
    SOMEJSON2 = '''
    [
        {
            "id":"01",
            "name": "Java",
            "thing":{},
            "list":[]
        },
        {
            "id":"07",
            "name": "C++",
            "thing":{},
            "list":[]
        },
        {
            "id":"08",
            "name": "DDD",
            "thing":{},
            "list":[],
            "extra":23
        }
    ]
    '''

    def test_domonic_JSON(self):
        # data = JSON.parse(SOMEJSON)
        # print(data)

        t = JSON.tablify(TestCase.SOMEJSON2)
        print(t)

        t = JSON.tablify(json.loads(TestCase.SOMEJSON2))
        print(t)

        t = JSON.tablify(TestCase.SOMEJSON)
        print(t)

        # t = JSON.csvify(TestCase.SOMEJSON)
        # print(t)

        # t = JSON.csv2json("data.csv")
        # print(t)

        @return_json
        def yo():
            myObj = {"hi": [1, 2, 3]}
            return myObj

        print('goat')
        print(yo())

        # json_data = JSON.parse_file('surveys.json')
        # mytable = JSON.tablify(json_data)
        # print(mytable)

        # with JSON( data, 'items') as item:
        # print(item)
        # print(item.id)

        # iterator = JSON( data, 'items.age', lambda i: i<30 )

        # @JSON
        # def somefunc():
        #     returns pyobj


if __name__ == '__main__':
    unittest.main()



"""
    domonic.i18n.ch
    ====================================

    html tags in Chinese

    first translated to Chinese using google translate.
    send a pull request if you have a better translation.
"""
from domonic.html import *

Shēntǐ = 身体 = body
tóu = 头 = head
jiǎoběn = 脚本 = script
fēnggé = 风格 = style
chuánsòngmén = 传送门 = portal
qiáng = 强 = strong
kuàiyǐnyòng = 块引用 = blockquote
biǎo = 表 = table
xíngshì = 形式 = form
biāoqiān = 标签 = label
tíjiāo = 提交 = submit
biāotí = 标题 = title
# noscript = noscript = noscript
bùfèn = 部分 = section
dǎoháng = 导航 = nav
wénzhāng = 文章 = article
yībiān = 一边 = aside
# hgroup = hgroup = hgroup
dìzhǐ = 地址 = address
shùzì = 数字 = figure
# figcaption = figcaption = figcaption
xiǎo = 小 = small
yǐnyòng = 引用 = cite
dàimǎ = 代码 = code
zi = 子 = sub
chāo = 超 = sup
biāojì = 标记 = mark
hóngbǎoshí = 红宝石 = ruby
kuàdù = 跨度 = span
# iframe = iframe = iframe
shìpín = 视频 = video
yīnpín = 音频 = audio
huàbù = 画布 = canvas
biāotí = 标题 = caption
zǔqún = 组群 = colgroup
shēntǐ = 身体 = tbody
# tóu = 头 = thead
jiǎo = 脚 = tfoot
zìduànjí = 字段集 = fieldset
chuánshuō = 传说 = legend
ànniǔ = 按钮 = button
xuǎnzé = 选择 = select
shùjùlièbiǎo = 数据列表 = datalist
xuǎnzézǔ = 选择组 = optgroup
xuǎnxiàng = 选项 = option
wénběnqūyù = 文本区域 = textarea
shūchū = 输出 = output
jìnbù = 进步 = progress
mǐ = 米 = meter
xiángqíng = 详情 = details
zǒngjié = 总结 = summary
càidān = 菜单 = menu
càidānxiàng = 菜单项 = menuitem
zìtǐ = 字体 = font
biāotí = 标题 = header
yèjiǎo = 页脚 = footer
jīshù = 基数 = base
liànjiē = 链接 = link
yuán = 元 = meta
túpiàn = 图片 = img
qiànrù = 嵌入 = embed
láiyuán = 来源 = source
guǐdào = 轨道 = track
miànjī = 面积 = area
shàngxiào = 上校 = col
shūrù = 输入 = input
mìnglìng = 命令 = command
zhǔyào = 主要 = main
jīběnzìtǐ = 基本字体 = basefont
zhōngxīn = 中心 = center
qiànrù = 嵌入 = embed
zìtǐ = 字体 = font
# isindex = isindex = isindex
lièbiǎo = 列表 = listing
míngwén = 明文 = plaintext
bàgōng = 罢工 = strike
múbǎn = 模板 = template
túpiàn = 图片 = picture
wéndànglèixíng = 文档类型 = doctype
pínglùn = 评论 = comment



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

from domonic import __version__
version = __version__

setup(
  name='domonic',
  version=version,
  author="@byteface",
  author_email="byteface@gmail.com",
  license="MIT",
  url='https://github.com/byteface/domonic',
  download_url='https://github.com/byteface/domonic/archive/' + version + ' .tar.gz',
  description='Generate html with python 3. DOM API, Javascript API and more...',
  long_description=long_description,
  long_description_content_type="text/markdown",
  keywords=['html', 'generate', 'templating', 'dom', 'vdom', 'terminal', 'json', 'web', 'template', 'javascript', 'DOM', 'GUI', 'render', 'website', 'apps', 'html5', 'framework', 'SVG', 'x3d', 'events', 'geom'],
  python_requires='>=3.6',
  classifiers=[
      "Programming Language :: Python :: 3",
      "Programming Language :: JavaScript",
      "Programming Language :: Python",
      "Programming Language :: Python :: 3.6",
      "Programming Language :: Python :: 3.7",
      "Programming Language :: Python :: 3.8",
      "Programming Language :: Python :: 3.9",
      "Programming Language :: Python :: 3.10",
      "Development Status :: 4 - Beta",
      "Environment :: Web Environment",
      "Intended Audience :: Developers",
      "Intended Audience :: Other Audience",
      "License :: OSI Approved :: MIT License",
      "Natural Language :: English",
      "Operating System :: OS Independent",
      "Topic :: Internet",
      "Topic :: Internet :: WWW/HTTP",
      "Topic :: Multimedia :: Graphics :: Presentation",
      "Topic :: Software Development",
      "Topic :: Software Development :: Code Generators",
      "Topic :: Terminals",
      "Topic :: Utilities",
      'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
      'Topic :: Software Development :: Libraries :: Python Modules',
      'Topic :: Text Processing :: Markup :: HTML',
  ],
  install_requires=[
      'requests==2.26.0', 'python-dateutil==2.8.2', 'urllib3==1.26.7'
  ],
  packages=find_packages(),
  include_package_data=True,
  entry_points={
      'console_scripts': [
          'domonic = domonic.__main__:run',
      ],
  },
)



"""
    domonic.webapi.storage
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Storage
"""

import os
import json


class Storage():

    def __init__(self, filepath=None):
        """[localstorage. destroys on each session unless you pass the optional filepath]

        Args:
            filepath ([type], optional): [filepath]. give us a file to write to
        """
        self.storage = {}
        self.has_file = False
        if filepath:
            self.filepath = filepath
            self.has_file = True
        # check if file exists. if so load it in . if not create it
        if filepath:
            if os.path.exists(filepath):
                with open(filepath, 'r') as f:
                    self.storage = json.load(f)
            else:
                with open(filepath, 'w') as f:
                    json.dump(self.storage, f)

    def __getitem__(self, key):
        return self.storage[key]

    def __setitem__(self, key, value):
        self.storage[key] = value
        if self.has_file:
            self._save()

    def __len__(self):
        return len(self.storage.keys())

    @property
    def length(self):
        """ Returns an integer representing the number of data items stored in the Storage object. """
        return len(self.storage.keys())

    def _save(self):
        if self.has_file:
            with open(self.filepath, 'w') as f:
                json.dump(self.storage, f)
            return True
        return False

    def setItem(self, keyName, keyValue):
        """ Adds that key to the storage, or update that key's value if it already exists """
        self.storage[keyName] = keyValue
        self._update_file()

    def getItem(self, keyName):
        """ Returns the value of the specified key name Storage """
        return self.storage.get(keyName, None)

    def key(self, keyName):
        """ Returns the value of the key if it exists, otherwise returns None """
        return self.storage.get(keyName, None)

    def removeItem(self, keyName):
        """ Removes the key and its value from the storage """
        if keyName in self.storage:
            del self.storage[keyName]
            self._update_file()

    def clear(self):
        """ Removes all items from the storage """
        self.storage = {}
